# will contain a series of tools to better handle dicts

import numpy as np
import json

def string_to_dict(string):
    """
    converts a string representation of a dict back to a real dict
    :param string:
    :return:
    """
    if string is None:
        return None
    return json.loads(string.replace("'", '"').replace('None', 'null'))

def invert_keys_and_values(input_dict, warn_on_dupes=True):
    """
    swaps keys and values of a dict
    :param input_dict:
    :param warn_on_dupes:
    :return:
    """
    output_dict = {value:key for key, value in input_dict.items()}
    if warn_on_dupes:
        if len(output_dict)!=len(input_dict):
            print('Warning: some values in the dict are duplicated, keys and values cannot be inverted safely')
    return output_dict

def dict_to_numpy(input_dict):
    """
    converts a dict to a numpy array
    :param input_dict:
    :return:
    """
    # TODO add controls and report if dupe entries for example (I could get len of dict and len of the array first dim to see if all is ok
    return np.array(list(input_dict.items()))

def get_values_in_dict_matching_a_search_list(input_dict, search_list, ignore_not_found=True, value_for_not_found_if_not_ignored=None):
    """
    can be used to find all the values that match a search in a dict, can be used for retromapping for example
    :param input_dict:
    :param search_list:
    :param ignore_not_found:
    :param value_for_not_found_if_not_ignored:
    :return:
    """
    if ignore_not_found:
        output = [input_dict[search] for search in search_list if search in input_dict.keys()] # do I need to specify keys ????
    else:
        output = [input_dict[search] if search in input_dict.keys() else value_for_not_found_if_not_ignored for search in
                  search_list]  # do I need to specify keys ????
    return output

# convert smthg that is not a dict but has two values to a dict --> TODO add controls
def _to_dict(input):
    dct = {}
    for k,v in input:
        dct[k]=v
    return dct


# zut le code est pas finalisé...

# USE a track object to store the click and its content...


# do a code to do smart track correction that avoids creating dupes
# checks for dupes, etc...
# can connect tracks and or swap tracks

# si les deux couleurs existent dans les deux images alors c'est du swap et faut swapper à partir de la dernière image sinon
# si chaque couleur existe dans une image et pas dans l'autre alors faire un connect track. Si le connect risque de dupliquer une cellule alors il faut faire un swap de la cellule puis connecter la track afin de ne pas dupliquer les cellules
# reflechir mais ça devrait etre bon
# existe t'il un moyen perenne de restaurer les tracks corrected ???? si oui lequel pr que le job ne soit pas perdu
# peut etre simplement faire un truc du genre update tracks et rajouter les cellules perdues de maniere aleatoire ? pas si simple en fait
# puis je aussi swapper plusieurs cellules en même temps ??? peut etre mais pas si simple
# comment faire un algo de swap intelligent


# need to store cell id or centroid or centre of area and the frame --> should be sufficient to do everything --> the most robust then detect the color of the other

# offer apply to all or just before or after the current image
# do the codes for that

# aussi faire un code pour reappliquer les tracks


from epyseg.img import RGB_to_int24, int24_to_RGB
from timeit import default_timer as timer
import matplotlib.pyplot as plt
from epyseg.img import Img
from epyseg.utils.loadlist import loadlist
from epyseg.ta.tracking.tools import smart_name_parser
from epyseg.tools.logger import TA_logger  # logging

logger = TA_logger()

def swap(img, id1, id2):
    idcs1 = img == id1
    idcs2 = img == id2
    img[idcs1], img[idcs2] = id2, id1
    return img


def assign_id(img, old_id, new_id):
    img[img == old_id] = new_id
    return img

def swap_tracks(lst, frame_of_first_connection, id_t0, id_t1, __preview_only=False):
    correct_track(lst, frame_of_first_connection, id_t0, id_t1, correction_mode='swap', __preview_only=__preview_only)


def connect_tracks(lst, frame_of_first_connection, id_t0, id_t1,__preview_only=False):
    correct_track(lst, frame_of_first_connection, id_t0, id_t1, correction_mode='connect', __preview_only=__preview_only)


def correct_track(lst, frame_of_first_connection, id_t0, id_t1, correction_mode='connect', __preview_only=False, early_stop=True):
    if frame_of_first_connection >= len(lst):
        logger.error('error wrong connection frame nb')
        return

        # first thing to do is to check that we are not going to duplicate cells --> need check that both are not there together in the same frame otherwise need swap cells before --> try that
    for l in range(frame_of_first_connection, len(lst)):
        # file_path_0 = images_to_analyze[l]
        # filename0_without_ext = os.path.splitext(file_path_0)[0]


        tracked_cell_path = smart_name_parser(lst[l], ordered_output='tracked_cells_resized.tif') # os.path.join(filename0_without_ext, )
        tracked_cells_t0 = RGB_to_int24(Img(tracked_cell_path))

        cellpos_1 = id_t0 in tracked_cells_t0
        cellpos_2 = id_t1 in tracked_cells_t0

        if not cellpos_1 and not cellpos_2:
            logger.info('cells not found --> ignoring track correction')
            continue

        if cellpos_1 and cellpos_2 and correction_mode == 'connect':
            logger.info('IDs are both present in the same image, to avoid track ID duplication the desired change will be ignored, please ensure the selected cells are not wimply swapped in the image')
            # alternatively I could do a swap but not sure this is wise
            break

        if not cellpos_1 and cellpos_2 or not cellpos_2 and cellpos_1 and correction_mode == 'swap':
            logger.info('missing cell at frame '+str(l)+' swapping ignored')
            if early_stop:
                logger.info('quitting track edition')
                break
            else:
                continue

        if cellpos_1 and cellpos_2 and correction_mode == 'swap': # or (cellpos_1 and cellpos_2 and correction_mode == 'connect')
            logger.info('swapped cell ids: ' + str(id_t0) + ' and ' + str(id_t1))
            # or consider this an error and report it and ignore
            tracked_cells_t0 = swap(tracked_cells_t0, id_t0, id_t1)
            if __preview_only:
                plt.imshow(int24_to_RGB(tracked_cells_t0))
                plt.show()
            else:
                Img(int24_to_RGB(tracked_cells_t0), dimensions='hwc').save(tracked_cell_path, mode='raw')
        else:
            if cellpos_1:
                logger.info('changed id ' + str(id_t0) + ' to ' + str(id_t1))
                # tracked_cells_t0[tracked_cells_t0 == id_t0] = id_t1
                tracked_cells_t0 = assign_id(tracked_cells_t0, id_t0, id_t1)
            elif cellpos_2:
                logger.info('changed id ' + str(id_t1) + ' to ' + str(id_t0))
                # tracked_cells_t0[tracked_cells_t0 == id_t1] = id_t0
                tracked_cells_t0 = assign_id(tracked_cells_t0, id_t1, id_t0)
            if __preview_only:
                plt.imshow(int24_to_RGB(tracked_cells_t0))
                plt.show()
            else:
                Img(int24_to_RGB(tracked_cells_t0), dimensions='hwc').save(tracked_cell_path, mode='raw')

if __name__ == "__main__":
    start_all = timer()

    # path = '/E/Sample_images/tracking_test'
    path = '/E/Sample_images/sample_images_PA/trash_test_mem/mini10_fake_swaps/liste.lst'
    # path = '/E/Sample_images/test_tracking_with_shift' # quick test of the algo
    #
    # images_to_analyze = os.listdir(path)
    # images_to_analyze = [os.path.join(path, f) for f in images_to_analyze if
    #                      os.path.isfile(os.path.join(path, f))]  # list only files and only if they end by tif
    # images_to_analyze = natsorted(images_to_analyze)

    images_to_analyze = loadlist(path)

    # then do amrutha's code watershed  imagendarray (2-D, 3-D, …) of integers--> cool --> will also work for 3D --> then need segmentation in 3D
    #
    #
    # for l in range(len(images_to_analyze)):
    #     # in fact I don't even need to reopen the mask file because the track file should suffice ???
    #
    #     file_path_0 = images_to_analyze[l]
    #     if not file_path_0.endswith('.tif'):
    #         continue
    #
    #
    #     # print('files', file_path_1, file_path_0)
    #
    #     filename0_without_ext = os.path.splitext(file_path_0)[0]
    #
    #     tracked_cells_t0 = Img(os.path.join(filename0_without_ext,        'tracked_cells_resized.tif')) # 455:455 + 128, 435:435 + 128 #455:455 + 256, 435:435 + 290 #768:1024,0:256 #0:128, 0:128 #500:500+128, 500:500+128
    #
    #
    #
    #
    #     # TODO always get the mask instead of any other files in order to have the minimal nb of files available
    #     # mask_t0 = Img(os.path.join(filename0_without_ext, 'handCorrection.tif'))[
    #     #     ..., 0]  # 455:455 + 128, 435:435 + 128 #455:455 + 256, 435:435 + 290 #768:1024,0:256 #0:128, 0:128 #500:500+128, 500:500+128
    #
    #
    #     # labels_t0 = measure.label(mask_t0, connectivity=1,
    #     #                           background=255)  # FOUR_CONNECTED # use bg = 255 to avoid having to invert the image --> a gain of time I think and could be applied more generally
    #
    #     height = tracked_cells_t0.shape[0]
    #     width = tracked_cells_t0.shape[1]
    #
    #
    #     # could maybe store the ID of duplicate cell as a reseed --> can be a pb with extruding cells but should most often be ok though
    #     #
    #
    #     labels_tracking_t0 = measure.label(tracked_cells_t0[..., 0], connectivity=1, background=255)
    #     rps_t0 = regionprops(labels_tracking_t0)
    #
    #
    #     centroids_t0 = []
    #     for region in rps_t0:
    #         # take regions with large enough areas
    #         centroids_t0.append(region.centroid)

    # try create a dupe to see if it fails -> it will and I should really avoid that

    # swap_tracks(images_to_analyze, 1, 0x0ef5dd, 0xa28ad6, __preview_only=True)  # both cells exist


    swap_tracks(images_to_analyze, 0, 0x0ef5dd, 0xFAFAFF, __preview_only=True)  # one cells exist

    # ça marche mais reflechir que je ne puisse pas creer des pbs avec ça
    # connect_tracks(images_to_analyze, 0, 0xffdd5d, 0x3b2cdb, __preview_only=True)  # both cells exist
    # connect_tracks(images_to_analyze, 0, 0xffdd5d, 0x0ef5dd, __preview_only=True)  # create a duplicate cell # --> breaking

    # codes are the same --> just keep one and

    # tt a l'air de marcher --> maybe add a flag for recursive --> in fact not...

    print('total time', timer() - start_all)



# this class should contain all the global settings
import os

# default_UI = 'pyqt6' # 'non existing'
default_UI = 'pyqt6' #'pyqt6' #'pyqt6'#'pyside6' #'pyside2' #'pyqt5' # 'non existing'
# force = False

def default_qtpy_UI():
    # {'pyqt5': 'PyQt5', 'pyside2': 'PySide2', 'pyqt6': 'PyQt6', 'pyside6': 'PySide6'}
    global default_UI
    if default_UI not in list_UIs():
        default_UI = list_UIs()[0]
    return default_UI

def list_UIs():
    from qtpy import API_NAMES
    # return {'pyqt5': 'PyQt5', 'pyside2': 'PySide2', 'pyqt6': 'PyQt6', 'pyside6': 'PySide6'}.keys()
    return list(API_NAMES.keys())

def set_UI(ignore_if_already_set=True, UI=default_UI):
    if ignore_if_already_set and 'QT_API' in os.environ:
        # print('QT_API set to:'+os.environ['QT_API'] + '--> ignoring')
        return
    # print('before',os.environ['QT_API'])
    # if force:
    #     if UI is None:
    #         UI=default_UI
    # raise Exception # never called --> so where is the fucking bug
    if UI is None:
        UI = default_qtpy_UI()
    os.environ['QT_API'] = UI
    # print('chosen UI', UI)
    try:
        from qtpy.QtWidgets import QWidget
    except:
        # print('UI not found --> rolling back to default')
        # just fort safety, if pyqt6 is not found --> get back to default
        del os.environ['QT_API']

def print_default_qtpy_UI_really_in_use():
    try:
        UI_defined = None
        from qtpy.QtCore import PYQT_VERSION_STR
        try:
            UI_defined =os.environ['QT_API']
        except:
            pass
        print(UI_defined +' v'+PYQT_VERSION_STR)
    except:
        pass

def force_qtpy_to_use_user_specified(bool=True):
    if bool:
        os.environ['FORCE_QT_API'] = '1'  # keep this line --> this is absoultely required to get the framework to use what I sepcify otherwise it switches by default
    else:
        try:
            del os.environ['FORCE_QT_API']
        except:
            pass

if __name__ == '__main__':
    # import os
    # os.environ['QT_API']='tata'
    # print(os.environ['QT_API'])

    # os.environ['QT_API']='pyqt5'

    # default_UI='test'
    # # print(default_qtpy_UI())
    # set_UI()
    # print(os.environ['QT_API'])
    # set_UI()
    # print(os.environ['QT_API'])
    # print(list_UIs())
    #
    # print_default_qtpy_UI_really_in_use()

    # os.environ['QT_API'] = 'pyqt5'
    set_UI()
    print_default_qtpy_UI_really_in_use()



# this will be a scalebar object --> will contain text and a rectangle
# can have a position
# can set color of bars etc
# it needs to be associated to an image in a way
# but should it be part of the image or its own object linked to the image...
# need get bounds
# TODO make use of alignment to compute position of the bar --> just need its size then align and pack in the order wanted

# ça a l'air de marcher...
# MEGA TODO need rescale bar according to figure zoom in the real figure --> need draw on the original image space and not on the painter area
# TODO check if there is a bug in svg export if dpi is incorrect in the length of the scalebar --> wrong conversion factor --> really need check!!!

from epyseg.figure.alignment import alignRight, alignLeft, alignTop, alignBottom, alignCenterH, alignCenterV, packY, \
    packX
from epyseg.draw.shapes.rect2d import Rect2D
# logger
from epyseg.draw.shapes.txt2d import TAText2D
from epyseg.tools.logger import TA_logger
logger = TA_logger()

class ScaleBar(Rect2D):

    # TODO need a scaling factor for the bar if image is itself scaled or ??? --> most likely yes need think about it
    # TODO handle bar color
    def __init__(self, bar_width_in_units=0, legend="", unit_to_pixel_conversion_factor=1, bar_height_in_px=3, **kwargs):
        super(ScaleBar, self).__init__()
        self.scale = 1
        self.bar = Rect2D(color=None, fill_color=0xFFFFFF, stroke=0)
        self.bar.setWidth(bar_width_in_units * unit_to_pixel_conversion_factor)
        self.bar.setHeight(bar_height_in_px)
        self.legend = None
        self.setLegend(legend)
        self.bar_width_in_units = bar_width_in_units
        self.unit_to_pixel_conversion_factor = unit_to_pixel_conversion_factor

    def setLegend(self, legend):
        if isinstance(legend, str):
            self.legend = TAText2D(legend)
        elif isinstance(legend, TAText2D):
            self.legend = legend
        else:
            self.legend = None

    def setBarWidth(self, bar_width_in_units):
        self.bar_width_in_units = bar_width_in_units

    def setConversionFactor(self, unit_to_pixel_conversion_factor):
        self.unit_to_pixel_conversion_factor = unit_to_pixel_conversion_factor

    def draw(self, painter, draw=True):
        self.bar.draw(painter=painter)
        if self.legend is not None:
            self.legend.draw(painter=painter)
        # self.drawAndFill(painter=painter)

    def updateBoudingRect(self):
        """
        updates the image bounding rect depending on content
        :return:
        """

        # contient deux objects et est donc la somme des deux
        # --> facile en theorie faire que du packing et de l'alignement à gauche droite ou ailleurs
        # pack in y direction the text and the bar

        x = None
        y = None
        x2 = None
        y2 = None

        to_pack = [self.bar, self.legend]
        for img in to_pack:
            if img is None:
                continue
            topLeft = img.get_P1()
            if x is None:
                x = topLeft.x()
            if y is None:
                y = topLeft.y()
            x = min(topLeft.x(), x)
            y = min(topLeft.y(), y)

            # print(img, img.boundingRect(), type(img))
            # print(img, img.boundingRect(), type(img), img.boundingRect().height())

            if x2 is None:
                x2 = topLeft.x() + img.boundingRect().width()
            if y2 is None:
                y2 = topLeft.y() + img.boundingRect().height()
            x2 = max(topLeft.x() + img.boundingRect().width(), x2)
            y2 = max(topLeft.y() + img.boundingRect().height(), y2)

        self.setX(x)
        self.setY(y)
        self.setWidth(x2 - x)
        self.setHeight(y2 - y)

    def packY(self, space=3):
        # center in vertical axis then pack
        #
        # last_x = 0
        # last_y = 0
        #
        # to_pack = [self.bar, self.legend]
        # for i in range(len(to_pack)):
        #     img = self.images[i]
        #     if i != 0:
        #         last_y += space
        #     img.set_P1(img.get_P1().x(), last_y)
        #     # get all the bounding boxes and pack them with desired space in between
        #     # get first point and last point in x
        #     x = img.boundingRect().x()
        #     y = img.boundingRect().y()
        #     last_x = img.boundingRect().x() + img.boundingRect().width()
        #     last_y = img.boundingRect().y() + img.boundingRect().height()
        if self.legend is not None:
            self.bar.setWidth(self.scale * self.bar_width_in_units * self.unit_to_pixel_conversion_factor)
            alignCenterH(self.legend, self.bar)
            packY(3, self.legend, self.bar)

        self.updateBoudingRect()

    # finally just need align right

    def set_P1(self, *args):
        self.packY()
        curP1 = self.get_P1()
        Rect2D.set_P1(self, *args)
        newP1 = self.get_P1()

        to_pack = [self.bar, self.legend]
        for img in to_pack:
            if img is not None:
                img.translate(newP1.x() - curP1.x(), newP1.y() - curP1.y())

        self.updateBoudingRect()

    def set_to_scale(self, scale):
        self.scale = scale

    # def setcolor

    # get the bounds of the bar with the associated text
    # center text on bar see how I do in EZF



from collections import Counter
import matplotlib.pyplot as plt
import numpy as np
from scipy import ndimage
from scipy.ndimage import distance_transform_edt, generate_binary_structure
from scipy.signal import convolve2d
from skimage.color import gray2rgb
from skimage.feature import peak_local_max
from skimage.segmentation import quickshift, watershed, find_boundaries
from skimage.morphology import remove_small_objects
from skimage.morphology import skeletonize
from skimage.measure import label, regionprops
from epyseg.img import Img


def get_optimized_mask2(img, sauvola_mask=None, use_quick_shift=False, __VISUAL_DEBUG=False,
                        __DEBUG=False, score_before_adding=False, return_seeds=False):

    final_image = None
    if use_quick_shift:
        rotations = [0, 2, 3]
        kernels = [1, 1.33, 1.66, 2]

        for kern in kernels:
            for rot in rotations:
                segments_quick = getQuickseg(img, nb_of_90_rotation=rot, kernel_size=kern)
                segments_quick = segments_quick.astype(np.uint8)
                if final_image is None:
                    final_image = segments_quick
                else:
                    final_image = final_image + segments_quick

        if __VISUAL_DEBUG:
            plt.imshow(final_image)
            plt.title("avg")
            plt.show()

        final_image[final_image < final_image.max()] = 0
        final_image[final_image >= final_image.max()] = 1

    if sauvola_mask is None:
        from epyseg.postprocess.edmshed import sauvola
        t = sauvola(img, min_threshold=0.1, window_size=25)

        img[img >= t] = 1
        img[img < t] = 0
    else:
        img = sauvola_mask
        img[img > 0] = 1

    if __DEBUG:
        Img(img, dimensions='hw').save('/home/aigouy/Bureau/trash/trash4/sauvola_mask.tif')

    raw_sauvola = img.copy()

    if use_quick_shift:
        final_image[raw_sauvola != 0] = 1
        if __VISUAL_DEBUG:
            plt.imshow(final_image)
            plt.show()

        if __DEBUG:
            Img(final_image.astype(np.uint8) * 255, dimensions='hw').save(
                '/home/aigouy/Bureau/trash/trash4/corrected_stuff.tif')

        final_image = ~remove_small_objects(~final_image.astype(bool), min_size=5, connectivity=1)
        final_image = skeletonize(final_image)

        if __DEBUG:
            Img(final_image.astype(np.uint8), dimensions='hw').save('/home/aigouy/Bureau/trash/trash4/skel_quick.tif')
        if __VISUAL_DEBUG:
            plt.imshow(final_image)
            plt.title("binary")
            plt.show()

        vertices_quick, cut_bonds_quick = split_into_vertices_and_bonds(final_image)
        if __DEBUG:
            Img(vertices_quick.astype(np.uint8), dimensions='hw').save(
                '/home/aigouy/Bureau/trash/trash4/vertices_quick.tif')
            Img(cut_bonds_quick.astype(np.uint8), dimensions='hw').save(
                '/home/aigouy/Bureau/trash/trash4/cut_bonds_quick.tif')

    img = skeletonize(img)
    if __DEBUG:
        Img(img, dimensions='hw').save('/home/aigouy/Bureau/trash/trash4/skel_sauvola_mask.tif')

    img = remove_small_objects(img, min_size=6, connectivity=2)
    if __DEBUG:
        Img(img, dimensions='hw').save('/home/aigouy/Bureau/trash/trash4/skel_sauvola_mask_deblobed.tif')

    if __VISUAL_DEBUG:
        plt.imshow(img)
        plt.show()

    image = img.copy()
    image = image.astype(np.uint8) * 255
    image = Img.invert(image)

    distance = distance_transform_edt(image)

    if __VISUAL_DEBUG:
        plt.imshow(distance)
        plt.show()

    # local_maxi = peak_local_max(distance, indices=False, # old code changed due to deprecation
    #                             footprint=np.ones((8, 8)),
    #                             labels=image)
    tmp = peak_local_max(distance,
                         # indices=False, #change due to deprecation:  The indices argument in skimage.feature.peak_local_max has been deprecated. Indices will always be returned. (#4752)
                         footprint=np.ones((8, 8)),
                         labels=image)
    local_maxi = np.zeros_like(distance, dtype=bool)
    local_maxi[tuple(tmp.T)] = True

    distance = -distance

    markers = ndimage.label(local_maxi, structure=generate_binary_structure(2, 2))[0]

    if __VISUAL_DEBUG:
        plt.imshow(markers)
        plt.show()

    if __DEBUG:
        Img(markers, dimensions='hw').save('/home/aigouy/Bureau/trash/trash4/markers_0.tif')

    labels = watershed(distance, markers, watershed_line=True)  # --> maybe implement that too
    labels[labels != 0] = 1
    labels[labels == 0] = 255
    labels[labels == 1] = 0
    labels[labels == 255] = 1

    if __VISUAL_DEBUG:
        plt.imshow(labels)
        plt.title('raw wshed')
        plt.show()

    if __DEBUG:
        Img(labels.astype(np.uint8), dimensions='hw').save('/home/aigouy/Bureau/trash/trash4/wshed_edm.tif')

    vertices_edm, cut_bonds_edm = split_into_vertices_and_bonds(labels)
    if __DEBUG:
        Img(vertices_edm.astype(np.uint8), dimensions='hw').save('/home/aigouy/Bureau/trash/trash4/vertices_edm.tif')
        Img(cut_bonds_edm.astype(np.uint8), dimensions='hw').save('/home/aigouy/Bureau/trash/trash4/cut_bonds_edm.tif')

    if use_quick_shift:

        if __DEBUG:
            Img(img, dimensions='hw').save('/home/aigouy/Bureau/trash/trash4/skel_sauvola_mask_deblobed.tif')
        unconnected = detect_unconnected_bonds(img)

        if __DEBUG:
            Img(unconnected, dimensions='hw').save('/home/aigouy/Bureau/trash/trash4/unconnected.tif')

        labels_quick = label(cut_bonds_quick, connectivity=2, background=0)
        labels_quick_vertices = label(vertices_quick, connectivity=2, background=0)
        props_labels_quick = regionprops(labels_quick)

        labels_pred = label(unconnected, connectivity=2, background=0)

        raw_sauvola = rescue_bonds(labels_pred, labels_quick, raw_sauvola, labels_quick_vertices,
                                       props_labels_quick, score_before_adding=score_before_adding)

        if __DEBUG:
            Img(raw_sauvola.astype(np.uint8), dimensions='hw').save(
                '/home/aigouy/Bureau/trash/trash4/corrected_bonds_sauvola.tif')

    labels_edm = label(cut_bonds_edm, connectivity=2, background=0)
    labels_edm_vertices = label(vertices_edm, connectivity=2, background=0)
    props_labels_edm = regionprops(labels_edm)

    img = skeletonize(raw_sauvola)
    if __DEBUG:
        Img(raw_sauvola, dimensions='hw').save('/home/aigouy/Bureau/trash/trash4/skel_sauvola_mask2.tif')
    unconnected = detect_unconnected_bonds(img)

    if __DEBUG:
        Img(unconnected, dimensions='hw').save('/home/aigouy/Bureau/trash/trash4/unconnected.tif')

    labels_pred = label(unconnected, connectivity=2, background=0)

    raw_sauvola = rescue_bonds(labels_pred, labels_edm, raw_sauvola, labels_edm_vertices, props_labels_edm,
                                   score_before_adding=score_before_adding)

    raw_sauvola = connect_unconnected(labels_pred, labels_edm, raw_sauvola, props_labels_edm,
                                          labels_edm_vertices)

    if return_seeds:
        return markers, labels, raw_sauvola

    return raw_sauvola


def split_into_vertices_and_bonds(skel):
    kernel = np.ones((3, 3))
    mask = convolve2d(skel, kernel, mode='same', fillvalue=1)

    mask[mask < 4] = 0
    mask[mask >= 4] = 1  # vertices

    mask = np.logical_and(mask, skel).astype(np.uint8)

    # bonds without vertices
    bonds_without_vertices = skel - mask

    return mask, bonds_without_vertices

def connect_unconnected(labels_pred, labels_quick, raw_pahansalkar, props_labels_quick, labels_quick_vertices):
    final_bonds_to_restore_because_they_connect_unconnected = []
    for region in regionprops(labels_pred):
        ids = []
        for coordinates in region.coords:
            for i in range(-2, 3, 1):
                for j in range(-2, 3, 1):
                    try:
                        if labels_quick[coordinates[0] + i, coordinates[1] + j] != 0:
                            ids.append(labels_quick[coordinates[0] + i, coordinates[1] + j])
                    except:
                        pass
        if ids:
            ids = list(dict.fromkeys(ids))
            for id in ids:
                final_bonds_to_restore_because_they_connect_unconnected.append(id)

    final_bonds_to_restore_because_they_connect_unconnected = Counter(
        final_bonds_to_restore_because_they_connect_unconnected)

    id_of_vertices_to_restore = []

    for key, count in final_bonds_to_restore_because_they_connect_unconnected.items():
        if count == 2:
            # we are treating the vx and really reconnecting it then we remove it
            for coordinates in props_labels_quick[key - 1].coords:
                raw_pahansalkar[coordinates[0], coordinates[1]] = 1
                for i in range(-1, 2, 1):
                    for j in range(-1, 2, 1):
                        try:
                            if labels_quick_vertices[coordinates[0] + i, coordinates[1] + j] != 0:
                                id_of_vertices_to_restore.append(
                                    labels_quick_vertices[coordinates[0] + i, coordinates[1] + j])
                        except:
                            pass
    if id_of_vertices_to_restore:
        id_of_vertices_to_restore = list(dict.fromkeys(id_of_vertices_to_restore))
        props_labels_vertices_quick = regionprops(labels_quick_vertices)
        for id in id_of_vertices_to_restore:
            for coordinates in props_labels_vertices_quick[id - 1].coords:
                raw_pahansalkar[coordinates[0], coordinates[1]] = 1
    return raw_pahansalkar

def rescue_bonds(labels_pred, labels_quick, raw_pahansalkar, labels_quick_vertices, props_labels_quick,
                 score_before_adding=False):
    minimum_scores = 0.5
    minimal_area_for_scoring = 4
    restore_vertices = True

    bonds_to_rescue = {}
    id_of_vertices_to_restore = [] # there was a bug here --> variable was local

    for region in regionprops(labels_pred):
        ids = []
        last_pos = None



        for coordinates in region.coords:
            for i in range(-2, 3, 1):
                for j in range(-2, 3, 1):
                    try:
                        if i == 0 and j == 0:
                            last_pos = (coordinates[0] + i, coordinates[1] + j)
                        if labels_quick[coordinates[0] + i, coordinates[1] + j] != 0:
                            ids.append(labels_quick[coordinates[0] + i, coordinates[1] + j])

                        if restore_vertices:
                            if labels_quick_vertices[coordinates[0] + i, coordinates[1] + j] != 0:
                                id_of_vertices_to_restore.append(
                                    labels_quick_vertices[coordinates[0] + i, coordinates[1] + j])
                    except:
                        pass

        if ids:
            ids = list(dict.fromkeys(ids))
            bonds_to_rescue[last_pos] = ids

    bonds_to_rescue_to_remove = []

    # redo connected or unconnected
    for key, ids in bonds_to_rescue.items():
        for id in ids:
            if score_before_adding and props_labels_quick[id - 1].area > minimal_area_for_scoring:
                count = 0
                total = 0
                for coordinates in props_labels_quick[id - 1].coords:
                    total += 1
                    if raw_pahansalkar[coordinates[0], coordinates[1]] != 0:
                        count += 1
                if total != 0:
                    score = count / total
                    if score < minimum_scores:
                        bonds_to_rescue_to_remove.append(id)
                        continue

            # we are treating the vx and really reconnecting it then we remove it
            labels_pred[key[0], key[1]] = 0

            for coordinates in props_labels_quick[id - 1].coords:
                raw_pahansalkar[coordinates[0], coordinates[1]] = 1
                for i in range(-1, 2, 1):
                    for j in range(-1, 2, 1):
                        try:
                            if labels_quick_vertices[coordinates[0] + i, coordinates[1] + j] != 0:
                                id_of_vertices_to_restore.append(
                                    labels_quick_vertices[coordinates[0] + i, coordinates[1] + j])
                        except:
                            pass
    # ça marche mais faut aussi restaurer les vertices de ces bonds --> à refaire --> pas trop dur je pense --> faire pareil pr autre --> recup
    if id_of_vertices_to_restore:
        id_of_vertices_to_restore = list(dict.fromkeys(id_of_vertices_to_restore))
        props_labels_vertices_quick = regionprops(labels_quick_vertices)
        for id in id_of_vertices_to_restore:
            for coordinates in props_labels_vertices_quick[id - 1].coords:
                raw_pahansalkar[coordinates[0], coordinates[1]] = 1

    return raw_pahansalkar


def detect_unconnected_bonds(skel):
    kernel = np.ones((3, 3))
    mask = convolve2d(skel, kernel, mode='same', fillvalue=1)

    mask[mask == 0] = 255
    mask[mask > 2] = 0
    mask[mask == 2] = 1

    mask = np.logical_and(skel, mask).astype(np.uint8)
    return mask

def getQuickseg(img, nb_of_90_rotation=0, kernel_size=2):
    if nb_of_90_rotation == 0:
        return find_boundaries(quickshift(gray2rgb(img), kernel_size=kernel_size, max_dist=6, ratio=3))
    else:
        return np.rot90(find_boundaries(
            quickshift(gray2rgb(np.rot90(img, nb_of_90_rotation)), kernel_size=kernel_size, max_dist=6, ratio=3)),
            4 - nb_of_90_rotation)


if __name__ == '__main__':
    from timeit import default_timer as timer

    start = timer()

    #
    # image = Img('/D/final_folder_scoring/20190924_ecadGFP_400nM20E_000.tif')
    #
    # print(image.shape)
    # print(image.has_c())

    # img = Img('D:/Dropbox/mini_test.tif').astype(float)
    # img = Img('D:/Dropbox/stuff_for_the_new_figure/old/predict_avg_hq_correction_ensemble_wshed/5.tif')[...,0].astype(float)
    # img = Img('D:/Dropbox/stuff_for_the_new_figure/old/predict_avg_hq_correction_ensemble_wshed/122.tif')[...,0].astype(float)
    # img = Img('D:/Dropbox/stuff_for_the_new_figure/old/predict_avg_hq_correction_ensemble_wshed/11.tif')[...,0].astype(float)
    # img = Img('D:/Dropbox/stuff_for_the_new_figure/old/predict_avg_hq_correction_ensemble_wshed/cellpose_img22.tif')[...,0].astype(float)
    # img = Img('D:/Dropbox/stuff_for_the_new_figure/old/predict_avg_hq_correction_ensemble_wshed/cellpose_img22.tif')[...,5].astype(float)
    # img = Img('D:/Dropbox/stuff_for_the_new_figure/old/predict_avg_hq_correction_ensemble_wshed/cellpose_img22.tif')[...,1].astype(float)
    # img = Img('D:/Dropbox/stuff_for_the_new_figure/old/predict_avg_hq_correction_ensemble_wshed/cellpose_img22.tif')[...,3].astype(float)
    # img = Img.invert(img)
    # img = Img('D:/Dropbox/stuff_for_the_new_figure/old/predict_avg_hq_correction_ensemble_wshed/focused_Series010.tif')[...,0].astype(float)
    # img = Img('D:/Dropbox/stuff_for_the_new_figure/old/predict_avg_hq_correction_ensemble_wshed/focused_Series194.tif')[...,0].astype(float)
    # img = Img('D:/Dropbox/stuff_for_the_new_figure/old/predict_avg_hq_correction_ensemble_wshed/Series019.tif')[...,0].astype(float)
    # img = Img('D:/Dropbox/stuff_for_the_new_figure/old/predict_avg_hq_correction_ensemble_wshed/image_plant_best-zoomed.tif')[...,0].astype(float)
    # img = Img('D:/Dropbox/stuff_for_the_new_figure/old/predict_avg_hq_correction_ensemble_wshed/100708_png06.tif')[...,0].astype(float)
    # img = Img('D:/Dropbox/stuff_for_the_new_figure/old/predict_avg_hq_correction_ensemble_wshed/MAX_160610_test_ocelli_ok_but_useless_cause_differs_a_lot_from_ommatidia.lif - test_visualization_head_ommatidia_32h_APF_ok_2.tif')[...,0].astype(float)
    # img = Img('D:/Dropbox/stuff_for_the_new_figure/old/predict_avg_hq_correction_ensemble_wshed/proj0016.tif')[...,0].astype(float)

    # img = Img('D:/Dropbox/mini_test.tif').astype(float)
    # img = Img('/D/final_folder_scoring/predict_avg_hq_correction_ensemble_wshed/5.tif')[...,0].astype(float)
    # img = Img('/D/final_folder_scoring/predict_avg_hq_correction_ensemble_wshed/5.tif')[...,1].astype(float)
    # img = Img('/D/final_folder_scoring/predict_avg_hq_correction_ensemble_wshed/5.tif')[...,2].astype(float)
    # img = Img('/D/final_folder_scoring/predict_avg_hq_correction_ensemble_wshed/122.tif')[...,0].astype(float)
    # img = Img('/D/final_folder_scoring/predict_avg_hq_correction_ensemble_wshed/11.tif')[...,0].astype(float)
    # img = Img('/D/final_folder_scoring/predict_avg_hq_correction_ensemble_wshed/11.tif')[...,1].astype(float)
    # img = Img('/D/final_folder_scoring/predict_avg_hq_correction_ensemble_wshed/11.tif')[...,2].astype(float)
    # img = Img('/D/final_folder_scoring/predict_avg_hq_correction_ensemble_wshed/cellpose_img22.tif')[...,0].astype(float)
    # img = Img('/D/final_folder_scoring/predict_avg_hq_correction_ensemble_wshed/cellpose_img22_bg_subtracted_ij.tif')[...,0].astype(float)
    # img = Img('/D/final_folder_scoring/predict_avg_hq_correction_ensemble_wshed/focused_Series010.tif')[...,0].astype(float)
    # img = Img('/D/final_folder_scoring/predict_avg_hq_correction_ensemble_wshed/focused_Series194.tif')[..., 0].astype(float)
    # img = Img('/D/final_folder_scoring/predict_avg_hq_correction_ensemble_wshed/focused_Series194.tif')[..., 1].astype(float)
    # img = Img('/D/final_folder_scoring/predict_avg_hq_correction_ensemble_wshed/focused_Series194.tif')[..., 2].astype(float)

    img = Img('/D/final_folder_scoring/predict/11.tif')[..., 0]
    raw_sauvola = get_optimized_mask2(img, __VISUAL_DEBUG=True, __DEBUG=True)
    Img(raw_sauvola.astype(np.uint8), dimensions='hw').save(
        '/home/aigouy/Bureau/trash/trash4/corrected_bonds_sauvola2.tif')

    print('total time', timer() - start)



# Here I will gather all the unit tests for GUIs
# make sure all runs with pyside, pyside2, pyqt5 and pyqt6
import unittest
import os
import sys
import warnings
import random

from epyseg.settings.global_settings import print_default_qtpy_UI_really_in_use, list_UIs, set_UI, \
    force_qtpy_to_use_user_specified

# it works but I had to deactivate some values --> see why and what is the pb with the deactivated one!!!

class TestDialogs(unittest.TestCase):

    # DEV COMMENT TO KEEP: uncomment the lines below to check one specific UI
    # UI = 'pyqt5'
    UI = 'pyqt6'
    # UI = 'pyside2'
    # UI = 'pyside6'

    dialogs = ['DenoiseRecursionDialog',
               'crop_or_preview',
               'DefineROI',
               'minisel',               # GreekSelectorDialog
               'image_input_settings',  # will that work ???
               'Multiple_inputs',  # will that work ???
               'OpenFileOrFolderWidget',
               'PostProcessGUI',
               'DenoiseRecursionDialog',
               'FinishAllDialog',
               'QPlainTextInputDialog',
               'TrackingDialog',
               'WshedDialog',
               'Example', # buggy too
               'SQL_column_name_and_type',
               'DataAugmentationGUI', # this one is buggy --> remove
               ]  # dialogs to test


    def test_dialogs(self):
        force_qtpy_to_use_user_specified()
        os.environ['QT_API']=self.UI
        from epyseg.ta.GUI.denoise_recursion_dialog import DenoiseRecursionDialog
        from epyseg.gui.preview import crop_or_preview
        from epyseg.gui.defineROI import DefineROI
        from epyseg.gui.img import image_input_settings
        from epyseg.gui.mini.miniGUI_selection_dialog import minisel
        from epyseg.gui.multi_inputs_img import Multiple_inputs
        from epyseg.gui.open import OpenFileOrFolderWidget
        from epyseg.postprocess.gui import PostProcessGUI
        from epyseg.ta.GUI.denoise_recursion_dialog import DenoiseRecursionDialog
        from epyseg.ta.GUI.finish_all_dialog import FinishAllDialog
        from epyseg.ta.GUI.input_text_dialog import QPlainTextInputDialog
        from epyseg.ta.GUI.tracking_static_dialog import TrackingDialog
        from epyseg.ta.GUI.watershed_dialog import WshedDialog
        from epyseg.ta.database.preview_and_edit_sqlite_db import Example # buggy
        from epyseg.ta.database.sql_column_name_dialog import SQL_column_name_and_type
        from epyseg.gui.augmenter import DataAugmentationGUI #creates a bug probably misses some parameters!!!!
        from qtpy.QtWidgets import QApplication
        app = QApplication(sys.argv)
        print_default_qtpy_UI_really_in_use()
        self.assertEqual(os.environ['QT_API'], self.UI)


        # debug
        # gui = self.dialogs[4] # 0,1,2 -> ok
        # w = locals()[gui]()
        # # w = DenoiseRecursionDialog()
        # w.show()
        # self.assertTrue(w.isVisible())

        #
        #
        for iii,gui in enumerate(self.dialogs):
            print(iii, gui)
            w = locals()[gui]()
            w.show()
            # result = w.exec_()
            self.assertTrue(w.isVisible())
            # if iii == 1:
            #     break
            # w.close()
        # # self.w.btn.click()
        #
        #
        # # self.assertTrue(DenoiseRecursionDialog.isVisible())
        #
        #     # if random.uniform(0, 1)<0.5:
        #     #     self.assertTrue(False)
        # self.assertEqual(os.environ['QT_API'], self.UI)

if __name__ == '__main__':
    # DEV COMMENT KEEP nothing below works --> must uncomment at the beginning of the code

    # import os
    # os.environ['QT_API']='pyqt5'
    # TestGUI.UI ='pyqt5'
    # TestGUI.UI ='pyqt6'
    # TestGUI.UI ='pyside6'
    # TestGUI.UI ='pyside2'
    # TestGUI.UI = os.environ.get('TEST_UI', TestGUI.UI)
    # ne marche pas... --> need change it directly at the tip of the code
    unittest.main()





# TODO recode all from scratch rather than waisting time fixing it
# do a row kind of element that aligns vertically


# ça marche aussi maintenant reste plus qu'à faire le set width stuff set to width
def sameHeight(things_to_align, space):

    print('len(things_to_align)', len(things_to_align))
    if space is None:
        space = 0
    max_height = None
    # try:
    #     iter(things_to_align)
    for img in things_to_align:

        if max_height is None:
            max_height = img.boundingRect().height()
        max_height = max(img.boundingRect().height(), max_height)
        print('TUTU', img, type(img))
        print('TUTA', img, img.boundingRect(), img.boundingRect().height)
        print('TOTO', img, img.boundingRect(), img.boundingRect().height())


    # why not looping till the end --> only gets one
    count = 0
    for img in things_to_align:
        print(len(things_to_align), count)
        count+=1
        print('setting to height', max_height)
        print('before 2', type(img))
        img.setToHeight(max_height)
        print('after 2', type(img), img.boundingRect())

    print('max_height', max_height)

    packX(things_to_align,space)
    alignTop(things_to_align,updateBounds=False)
    # things_to_align.updateBoudingRect()

def packX(things_to_align, space=3):
    last_x = 0
    last_y = 0

    for i in range(len(things_to_align)):
        img = things_to_align[i]
        if i != 0:
            last_x += space
        img.set_P1(last_x, img.get_P1().y())
        last_x = img.boundingRect().x() + img.boundingRect().width()

    # things_to_align.updateBoudingRect()

def packY(things_to_align, space=3):
    last_x = 0
    last_y = 0


    print('1',things_to_align)


    for i in range(len(things_to_align)):
        img = things_to_align[i]
        print('2',img)
        if isinstance(img, list):
            from epyseg.draw.widgets.image_group import group
            img = group(*img)
        # should I create a row of objects in fact I need a group and move everything with that --> TODO --> maybe
        if i != 0:
            last_y += space

        print('img', type(img))

        img.set_P1(img.get_P1().x(), last_y)
        # get all the bounding boxes and pack them with desired space in between
        # get first point and last point in x
        x = img.boundingRect().x()
        y = img.boundingRect().y()
        last_x = img.boundingRect().x() + img.boundingRect().width()
        last_y = img.boundingRect().y() + img.boundingRect().height()
    # things_to_align.updateBoudingRect()



# Align vectorial objects to the top
# in fact should align in y
def alignTop(things_to_align, updateBounds=False):
    first_left = None
    for img in things_to_align:
        cur_pos = img.get_P1()
        if first_left is None:
            first_left = cur_pos
        img.set_P1(cur_pos.x(), first_left.y())
    # if updateBounds:
    #     things_to_align.updateMasterRect()



import os
from epyseg.settings.global_settings import set_UI # set the UI to be used py qtpy
set_UI()
#import argparse # later maybe allow arguments
from epyseg.epygui import EPySeg
from qtpy.QtWidgets import QApplication
import sys

if __name__ == '__main__':
    # run the deep learning GUI
    # code can be run from the command line using 'python -m epyseg'
    app = QApplication(sys.argv)
    w = EPySeg()
    w.show()
    sys.exit(app.exec_())




import os
from epyseg.settings.global_settings import set_UI # set the UI to be used py qtpy
set_UI()
from qtpy.QtCore import QRect, Qt, QRectF
from qtpy.QtWidgets import QWidget, QApplication, QGridLayout, QScrollArea
from epyseg.draw.shapes.square2d import Square2D
from epyseg.draw.widgets.paint import Createpaintwidget
from epyseg.img import Img
from epyseg.draw.shapes.rect2d import Rect2D
import sys

# in fact that is maybe already what I want!!!
# but I may also want to draw on it with a pen --> should have everything

class crop_or_preview(QWidget):

    def __init__(self, parent_window=None, preview_only=False):
        super().__init__(parent=parent_window)
        self.scale = 1.0
        self.x1 = self.x2 = self.y1 = self.y2 = None
        self.preview_only = preview_only
        self.initUI()

    def initUI(self):
        layout = QGridLayout()
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)

        self.paint = Createpaintwidget()

        self.paint.vdp.active = True
        self.paint.vdp.drawing_mode = True
        if not self.preview_only:
            self.paint.vdp.shape_to_draw = Rect2D

        self.scrollArea = QScrollArea()
        self.scrollArea.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.scrollArea.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)

        self.scrollArea.setWidget(self.paint)
        self.paint.scrollArea = self.scrollArea
        self.setMouseTracking(not self.preview_only)
        self.paint.setMouseTracking(not self.preview_only)  # KEEP IMPORTANT
        self.paint.mouseMoveEvent = self.mouseMoveEvent
        self.paint.mousePressEvent = self.mousePressEvent
        self.paint.mouseReleaseEvent = self.mouseReleaseEvent

        self.prev_width = 192
        self.prev_height = 192

        self.scrollArea.setGeometry(QRect(0, 0, self.prev_width, self.prev_height))
        self.setGeometry(QRect(0, 0, self.prev_width, self.prev_height))
        self.setFixedSize(self.size())
        layout.addWidget(self.scrollArea)

        self.setLayout(layout)

    def set_image(self, img):
        self.paint.vdp.shapes.clear()
        self.paint.setImage(img)  # bug is here
        if img is None:
            self.paint.scale = self.scale = self.paint.vdp.scale = 1.
        else:
            max_size = min(self.prev_width / img.get_width(), self.prev_height / img.get_height())
            self.paint.scale = self.scale = self.paint.vdp.scale = max_size

        if self.paint.image is not None:
            self.paint.resize(self.scale * self.paint.image.size())
            self.scrollArea.resize(self.scale * self.paint.image.size())

    def mousePressEvent(self, event):
        self.paint.vdp.shapes.clear()
        if self.paint.vdp.active:
            self.paint.vdp.mousePressEvent(event)
            if self.paint.vdp.currently_drawn_shape is not None:
                self.paint.vdp.currently_drawn_shape.stroke = 3 / self.scale
            self.update()

    def mouseMoveEvent(self, event):
        self.paint.vdp.mouseMoveEvent(event)
        region = self.scrollArea.widget().visibleRegion()
        self.paint.update(region)

    def mouseReleaseEvent(self, event):
        if self.paint.vdp.active:
            try:
                self.paint.vdp.mouseReleaseEvent(event)
                self.update()  # required to update drawing
                self.update_ROI()
            except:
                pass

    def set_square_ROI(self, bool):
        if bool:
            self.paint.vdp.shape_to_draw = Square2D
        else:
            self.paint.vdp.shape_to_draw = Rect2D

    def setRoi(self, x1, y1, x2, y2):
        # if x1 is None and y1 is None and x2 is None and y2 is None:
        #     self.paint.vdp.shapes.clear()
        #     self.paint.update()
        #     self.update()  # required to update drawing
        #     self.update_ROI()
        #     return
        if x1 != x2 and y1 != y2:
            # TODO add controls for size of ROI --> TODO but ok for now

            self.paint.vdp.shapes.clear()
            if x1 is not None and y1 is not None:
                rect2d = Rect2D(x1, y1, x2 - x1, y2 - y1)
                rect2d.stroke = 3 / self.scale
                self.paint.vdp.shapes.append(rect2d)
                self.paint.vdp.currently_drawn_shape = None
                self.paint.update()
                self.update()  # required to update drawing
                self.update_ROI()
            else:
                self.x1 = x1
                self.x2 = x2
                self.y1 = y1
                self.y2 = y2
                self.paint.update()
                self.update()

            # self.paint.vdp.update()
            # self.paint.vdp.currently_drawn_shape.stroke = 3 / self.scale
            # self.x1 = x1
            # self.x2 = x2
            # self.y1 = y1
            # self.y2 = y2


    def update_ROI(self):
        try:
            rect = self.paint.vdp.shapes[0]
            x1 = rect.x()
            y1 = rect.y()
            x2 = rect.x() + rect.width()
            y2 = rect.y() + rect.height()

            if x1 < 0:
                x1 = 0
            if y1 < 0:
                y1 = 0

            if x2 < 0:
                x2 = 0
            if y2 < 0:
                y2 = 0

            if rect.width() >= self.paint.image.size().width():
                x2 = self.paint.image.size().width()
            if rect.height() >= self.paint.image.size().height():
                y2 = self.paint.image.size().height()

            if x1 > x2:
                tmp = x2
                x2 = x1
                x1 = tmp

            if y1 > y2:
                tmp = y2
                y2 = y1
                y1 = tmp

            if x1 == x2:
                x1 = x2 = None
            if y1 == y2:
                y1 = y2 = None

            self.x1 = x1
            self.x2 = x2
            self.y1 = y1
            self.y2 = y2
        except:
            self.x1 = self.x2 = self.y1 = self.y2 = None

    def get_crop_parameters(self):
        if self.paint.vdp.shapes:
            self.update_ROI()
        if self.x1 is None:
            if self.x2 is not None:
                return {'x1': self.x1, 'y1': self.y1, 'w': int(self.x2), 'h': int(self.y2)}
            else:
                return None
        return {'x1': int(self.x1), 'y1': int(self.y1), 'x2': int(self.x2), 'y2': int(self.y2)}

if __name__ == '__main__':

    # ok in  fact that is already a great popup window --> can I further improve it ???

    # just for a test
    app = QApplication(sys.argv)
    ex = crop_or_preview()
    # ex = crop_or_preview(preview_only=True)
    # img = Img('/home/aigouy/mon_prog/Python/Deep_learning/unet/data/membrane/test/11.png')
    # img = Img('/home/aigouy/mon_prog/Python/Deep_learning/unet/data/membrane/test/122.png')
    # img = Img('/home/aigouy/mon_prog/Python/data/3D_bicolor_ovipo.tif')
    # img = Img('/home/aigouy/mon_prog/Python/data/Image11.lsm')
    # img = Img('/home/aigouy/mon_prog/Python/data/lion.jpeg')
    # img = Img('/home/aigouy/mon_prog/Python/data/epi_test.png')
    img = Img('/E/Sample_images/sample_images_PA/trash_test_mem/mini10_fake_swaps/focused_Series012.png')
    ex.set_image(img)

    # test = QRectF(None, None, 128, 128)

    ex.setRoi(None, None, 128, 256)
    # ex.set_image(None)
    ex.show()
    app.exec_()

    print(ex.get_crop_parameters())



import os
from epyseg.settings.global_settings import set_UI # set the UI to be used py qtpy
set_UI()
from qtpy.QtCore import QRect
from qtpy.QtWidgets import QApplication, QGridLayout, QLabel, QSpinBox, QDialog, QDialogButtonBox, QCheckBox, QFrame
import sys
from epyseg.tools.logger import TA_logger # logging

logger = TA_logger()


class DefineROI(QDialog):

    def __init__(self, parent_window=None, x1=0, y1=0, x2=0, y2=0):
        super().__init__(parent=parent_window)
        self.setWindowTitle('Define your ROI')
        self.x1 = x1
        self.x2 = x2
        self.y1 = y1
        self.y2 = y2
        self.initUI()

    def initUI(self):
        layout = QGridLayout()
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setColumnStretch(0, 25)
        layout.setColumnStretch(1, 75)

        self.random_checkbox = QCheckBox('Random ROI (random crop with defined width and height)', objectName='random_checkbox')
        self.random_checkbox.setChecked(False)

        labelX1 = QLabel('x1')
        self.x1_spinner = QSpinBox(objectName='x1_spinner')
        self.x1_spinner.setSingleStep(1)
        self.x1_spinner.setRange(0, 1_000_000)
        if self.x1 is not None:
            self.x1_spinner.setValue(self.x1)
        else:
            self.random_checkbox.setChecked(True)

        labelY1 = QLabel('y1')
        self.y1_spinner = QSpinBox(objectName='y1_spinner')
        self.y1_spinner.setSingleStep(1)
        self.y1_spinner.setRange(0, 1_000_000)
        if self.y1 is not None:
            self.y1_spinner.setValue(self.y1)
        else:
            self.random_checkbox.setChecked(True)

        self.labelX2 = QLabel('x2')
        self.x2_spinner = QSpinBox(objectName='x2_spinner')
        self.x2_spinner.setSingleStep(1)
        self.x2_spinner.setRange(0, 1_000_000)
        self.x2_spinner.setValue(self.x2)

        self.labelY2 = QLabel('y2')
        self.y2_spinner = QSpinBox(objectName='y2_spinner')
        self.y2_spinner.setSingleStep(1)
        self.y2_spinner.setRange(0, 1_000_000)
        self.y2_spinner.setValue(self.y2)

        self.random_checkbox.stateChanged.connect(self._change_mode)
        if self.random_checkbox.isChecked():
            self._change_mode()

        # line separator
        line_sep = QFrame()
        line_sep.setFrameShape(QFrame.HLine)
        line_sep.setFrameShadow(QFrame.Sunken)

        # self.setGeometry(QRect(0, 0, self.prev_width, self.prev_height))
        # self.setFixedSize(self.size())

        layout.addWidget(self.random_checkbox, 0, 0, 1, 2)
        layout.addWidget(labelX1, 1, 0)
        layout.addWidget(self.x1_spinner, 1, 1)
        layout.addWidget(labelY1, 2, 0)
        layout.addWidget(self.y1_spinner, 2, 1)
        layout.addWidget(self.labelX2, 3, 0)
        layout.addWidget(self.x2_spinner, 3, 1)
        layout.addWidget(self.labelY2, 4, 0)
        layout.addWidget(self.y2_spinner, 4, 1)
        layout.addWidget(line_sep, 5, 0, 1, 2)

        self.setLayout(layout)

        self.buttonBox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)

        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        layout.addWidget(self.buttonBox, 6, 0, 1, 2)

        self.setGeometry(QRect(0, 0, 300, 150))
        # self.setFixedSize(self.size())

    def _change_mode(self):
        if self.random_checkbox.isChecked():
            self.labelX2.setText('width')
            self.labelY2.setText('height')
            self.x1_spinner.setEnabled(False)
            self.y1_spinner.setEnabled(False)
        else:
            self.labelX2.setText('x2')
            self.labelY2.setText('y2')
            self.x1_spinner.setEnabled(True)
            self.y1_spinner.setEnabled(True)

    def getCrop(self):
        # TODO maybe add controls and return only if valid
        # Should I sort things too
        if self.x1_spinner.isEnabled():
            x1 = self.x1_spinner.value()
        else:
            x1 = None
        if self.y1_spinner.isEnabled():
            y1 = self.y1_spinner.value()
        else:
            y1 = None
        x2 = self.x2_spinner.value()
        y2 = self.y2_spinner.value()

        if x1 == x2:
            logger.error('x1=x2, ROI width is null --> ignoring ROI')
            return None
        if y1 == y2:
            logger.error('y1=y2, ROI height is null --> ignoring ROI')
            return None

        if x1 is not None:
            if x1 > x2:
                x1, x2 = x2, x1
        else:
            if x2 == 0:
                logger.error('ROI width is null --> ignoring ROI')
                return None
        if y1 is not None:
            if y1 > y2:
                y1, y2 = y2, y1
        else:
            if y2 == 0:
                logger.error('ROI height is null --> ignoring ROI')
                return None

        return x1, y1, x2, y2

    @staticmethod
    def getDataAndParameters(parent_window=None, x1=0, y1=0, x2=0, y2=0):  #
        dialog = DefineROI(parent_window=parent_window, x1=x1, y1=y1, x2=x2, y2=y2)

        result = dialog.exec_()
        if result:
            roi = dialog.getCrop()
        else:
            return None, False
        return (roi, result == QDialog.Accepted)


if __name__ == '__main__':
    # just for a test
    app = QApplication(sys.argv)
    # ex = DefineROI(x1=120, y1=26, x2=16, y2=300)
    # ex = DefineROI(x1=120, y1=26, x2=120, y2=300)
    # ex = crop_or_preview(preview_only=True)
    # img = Img('/home/aigouy/mon_prog/Python/Deep_learning/unet/data/membrane/test/11.png')
    # img = Img('/home/aigouy/mon_prog/Python/Deep_learning/unet/data/membrane/test/122.png')
    # img = Img('/home/aigouy/mon_prog/Python/data/3D_bicolor_ovipo.tif')
    # img = Img('/home/aigouy/mon_prog/Python/data/Image11.lsm')
    # img = Img('/home/aigouy/mon_prog/Python/data/lion.jpeg')
    # img = Img('/home/aigouy/mon_prog/Python/data/epi_test.png')
    # ex.set_image(None)

    # ROI, ok = DefineROI.getDataAndParameters(x1=120, y1=26, x2=120, y2=300)
    ROI, ok = DefineROI.getDataAndParameters(x1=None, y1=None, x2=128, y2=256)

    if ok:
        print(ROI)

    # ex.show()
    # app.exec_()
    sys.exit(0)



# this class will be used to triangulate a cell in order to compute its 3D properties such as area
import random

from scipy.spatial import Delaunay
from skimage.draw import polygon_perimeter

from epyseg.img import Img, RGB_to_int24

from epyseg.ta.measurements.measurements3D.get_point_on_surface_if_centroid_is_bad import point_on_surface
from epyseg.ta.measurements.measurements3D.measures3D import compute_3D_surfacearea
from epyseg.ta.tracking.rapid_detection_of_vertices import detect_vertices_and_bonds, associate_vertices_to_cells
import numpy as np
import matplotlib.pyplot as plt
from skimage.measure import regionprops
from skimage import measure
import traceback


# print(unprocessed_clone.shape)
# labels = label(RGB_to_int24(unprocessed_clone), connectivity=1, background=0xFFFFFF)
# vertices = detect_vertices_and_bonds(labels)

# check what I really need and add it there
# instead of region I could have the coords of the image

# is there a way to make it generic for cells and for clones --> think about it --> probably yes by just using the coords
# maybe check there is a single ID below
#
# coords_etdited = region.coords.tolist()
# coords_etdited.append(region.centroid)
#
# verts = np.where((vertices == 255) & (
#             clone_id == cell.label))  # if image has too many labels then use that if there are several clones inside the stuff
#
# tmpsss = np.zeros_like(labels)
# tmpsss[verts] = 255
# plt.imshow(tmpsss)
# plt.show()
#
# verts = list(zip(verts[0], verts[1]))
# # print(np.where(vertices == 255))
# coords_etdited.extend(verts)
# # coords_etdited = verts
#
# coords_etdited = np.asarray(coords_etdited)

# this is for filtering of the coords --> I could keep a light version of this in fact --> just make it generic

# from shapely.geometry import Polygon
#
# for triangle in triangles:
#     centroid_y, centroid_x = sum(triangle) / len(triangle)
#
#     # print('me', centroid_y, centroid_x)
#     if clone_id[int(centroid_y), int(centroid_x)] == cell.label:
#         poly = Polygon(triangle)
#
#         plt.plot(*poly.exterior.xy)
#
#         # parfait --> maintenant voir comment faire ça mieux et si je peux utiliser shapely ou pas ???
#
# # figure()
# # axis('equal')
# # plot(coords_etdited[:, 0], coords_etdited[:, 1], '.')
# # for i, j in edges:
# #     plot(coords_etdited[[i, j], 0], coords_etdited[[i, j], 1])
# # plt.triplot(coords_etdited[:, 0], coords_etdited[:, 1], tris)
# plt.show()

# labels, vertices,
def triangulate(coords_to_triangulate, triangulated_object_ID=None, labels_to_check=None):
    # TODO export all these triangulations so that they are more simply done
    # nb triangulation would be simpler if using just the vertices instead of the full contour --> maybe rely on vertices also for the edges and that is simple to do in a way
    # all this triangulation is only useful provided the image is 3D/has heightmap because otherwise the metrics are much simpler to handle
    # print('oubsi qsdsqdqsdqsd q')

    # this is slow here so it must be done once and much before in fact!!!

    # ça marche pas non plus faut vraiment plus de points

    # plt.imshow(cells2)
    # plt.show()

    # plt.imshow(vertices)
    # plt.show()
    # ça marche --> essayer d'ajouter des vertices

    tri = Delaunay(coords_to_triangulate)
    triangles = coords_to_triangulate[tri.vertices]

    # TODO also need compute the 3D area of the shape

    # print(triangles)

    # Polygon([(3.0, 0.0), (2.0, 0.0), (2.0, 0.75), (2.5, 0.75), (2.5, 0.6), (2.25, 0.6), (2.25, 0.2), (3.0, 0.2),
    #          (3.0, 0.0)])

    # ça marche par contre bug si inner hole --> faudrait faire qq chose de plus smart en fait
    # --> en fait devrait pas arriver --> en fait just warner l'utilisateur
    # sinon checker les triangles par rapport à l'image originale en fait c'est facile dans mon cas les points des interbonds doivent se trouver dans le clone --> pas trop dur en fait
    # mieux le centroid du triangle doit se trouver dans le clone et ça c'est vraimet super facile à faire et en plus j'ai pas besoin de truc compliques --> 100 % codable en python et TA compat à 100%

    # indices_to_delete

    # TODO check this code and verify how generic it is!!!
    if labels_to_check is not None and triangulated_object_ID is not None:
        # print('entering delete phase', triangles.shape)
        indices_to_delete = []
        for iii, triangle in enumerate(triangles):
            centroid_y, centroid_x = sum(triangle) / len(triangle)

            # print('me', centroid_y, centroid_x)
            # can I do it more smartly ??? using another ID
            if labels_to_check[int(centroid_y), int(centroid_x)] != triangulated_object_ID:
                # poly = Polygon(triangle)

                # plt.plot(*poly.exterior.xy)

                # parfait --> maintenant voir comment faire ça mieux et si je peux utiliser shapely ou pas ???

                # can I delete it
                indices_to_delete.append(iii)

        # print('indices_to_delete',indices_to_delete) # --> ok --> seems to work
        # ça marche bien ça --> à réutiliser!!!
        triangles = np.delete(triangles, indices_to_delete, axis=0)

        # print('after delete', triangles.shape)
    # check if that work

    return triangles


# if image is not plot as shapely if not then plot on the image --> very good idea
# TODO allow maybe fill and chose a color or generate random colors as in TA !!! --> would be a very good idea and that would make my life easier!!!
def plot_triangulation(triangles, img_to_plot_triangles_over=None):
    if img_to_plot_triangles_over is None:
        from shapely.geometry import Polygon
        for triangle in triangles:
            poly = Polygon(triangle)
            plt.plot(*poly.exterior.xy)
            plt.axis('equal')
        plt.show()
    else:
        for triangle in triangles:
            r = triangle[:, 0]  # get the y coords of the simplex
            c = triangle[:, 1]  # get the x coords of the simplex

            # print(np.asarray(r))
            # print(c)

            # TODO maybe plot the poygon and its contour too
            rr, cc = polygon_perimeter(r, c, shape=img_to_plot_triangles_over.shape)

            # print(rr, cc)
            #
            # print(type(rr), print(rr.dtype), rr.shape)
            # print(rr)
            img_to_plot_triangles_over[rr, cc] = random.randint(128,255) # not bad
        return img_to_plot_triangles_over

if __name__ == '__main__':


    # everything is fine --> just need finalize the thing and also measure the area and things like that
    # also find a way to order pixels and get the perimeter right
    # si > sqrt 2 alors need a connection --> put this as a section and try connect to first or last of every list --> maybe not so hard to implement and in the end all the points should be connected but try on a real example !!!


    # TODO !!!

    # do several tests
    # do tests with classical cells and with clones just to see how it behaves
    if True:
        # cell triangulation test
        # get any image with cells get vertices for cells
        # get centroid if inside the cell or a random point
        # launch the triangulation and display it

        CHECK_CONCAVE_CELLS_TRIANGULATION = True # maybe not really necessary for epithelial cells but really key for clones

        cells = Img('/E/Sample_images/sample_images_PA/trash_test_mem/mini/focused_Series012/tracked_cells_resized.png')
        cells = RGB_to_int24(cells)
        vertices = detect_vertices_and_bonds(cells)
        vertices = np.where(vertices == 255)
        t0_cells_n_vertices = associate_vertices_to_cells(cells, vertices, forbidden_colors=(0, 0xFFFFFF))
        # now for every cell do the triangulation
        # --> TODO
        labels = measure.label(cells, connectivity=1, background=0xFFFFFF)  # FOUR_CONNECTED
        # plt.imshow(labels)
        # plt.show()
        counter = 0
        out = np.zeros_like(cells)

        for region in regionprops(labels):
            # region.bbox
            # counter += 1
            # if counter == 60:
                cell_ID = cells[region.coords[0][0], region.coords[0][1]]
                # region.coords
                # print(type(np.asarray(t0_cells_n_vertices[cell_ID])))
                # print(type(region.coords))

                try:

                    points_for_triangulation = t0_cells_n_vertices[cell_ID] # why not all the cells are also entered in there --> do I have a bug and if so where ???
                    # if centroid is in the cell --> add it to the cells and that would be ok!!! otherwise take a random point in the cell
                    points_for_triangulation.append(point_on_surface(region, labels))
                except:
                    traceback.print_exc()
                    print('no vertices detected for this cell! --> need fix the algorithm')
                    # --> indeed it fails at detecting cells at the edges!!! -> need fix that --> see how I was doing that in TA and do the same here
                    continue


                try:
                    if not CHECK_CONCAVE_CELLS_TRIANGULATION:
                        triangles = triangulate(np.asarray(points_for_triangulation))
                    else:
                        triangles = triangulate(np.asarray(points_for_triangulation),triangulated_object_ID=cell_ID, labels_to_check=cells)

                    plot_triangulation(triangles, out)
                except:
                    traceback.print_exc()
                    print('failed for cell', region.area,
                      region.bbox)  # area 64 but not enough vertices --> I probably have a bug in my vertices code --> TODO --> fix that --> most likely at the edges
                    # TODO fix the algo

                # print(triangles)
                # break
        plt.imshow(out)
        plt.show()



    if True:

        from personal.pyTA.clones.clone_fuser import get_clone_contour
        heightmap = None # but for a real case this must be done

        # triangulate a clone
        # clone = Img('/E/Sample_images/sample_images_PA/trash_test_mem/mini/focused_Series012/tracked_clone.tif')
        # clone = Img('/E/Sample_images/sample_images_PA/trash_test_mem/mini/focused_Series012/tracked_clone_003.png')
        # clone = Img('/E/Sample_images/sample_images_PA/trash_test_mem/mini/focused_Series012/tracked_clone.png')
        clone = Img('/E/Sample_images/sample_images_PA/trash_test_mem/mini/focused_Series012/tracked_clone_001.png')
        # ça marche super --> mettre ce code de partout!!!

        clone_id, clone_contour = get_clone_contour(fused_clone=None, unprocessed_clone=clone,
                                                    return_clone_id=True)


        # print(clone.shape)
        # nb can I use that also for clones ???

        # both are ok
        if False:
            plt.imshow(clone_id)
            plt.show()

            plt.imshow(clone_contour)
            plt.show()

        clone = RGB_to_int24(clone)
        cell_labels = measure.label(clone, connectivity=1, background=0xFFFFFF)
        #
        vertices = detect_vertices_and_bonds(cell_labels)

        # ce truc est ok
        if False:
            plt.imshow(vertices)
            plt.show()

        # why no cells here --> see what the pb is I must have a bug in my code somewhere though

        # du coup ça ne marche pas ici
        t0_cells_n_vertices = associate_vertices_to_cells(clone_id, vertices, forbidden_colors=(0, 0xFFFFFF))

        # print(t0_cells_n_vertices) # pourtant c'est bon --> j'ai une tonne de vx pr la cellule --> pkoi pas trouvée alors

        # get the clone and try to triangulate it, then I'll be done !!!


        # labels = measure.label(clone, connectivity=1, background=0xFFFFFF)  # FOUR_CONNECTED
        out = np.zeros_like(clone_id)

        for region in regionprops(clone_id):
                # region.bbox
                # counter += 1
                # if counter == 60:
                cell_ID = clone_id[region.coords[0][0], region.coords[0][1]]
                # region.coords
                # print(type(np.asarray(t0_cells_n_vertices[cell_ID])))
                # print(type(region.coords))

                try:

                    points_for_triangulation = t0_cells_n_vertices[
                        cell_ID]  # why not all the cells are also entered in there --> do I have a bug and if so where ???
                    # if centroid is in the cell --> add it to the cells and that would be ok!!! otherwise take a random point in the cell
                    # here that may even be smart to add the centoid of all cells contained in the clone --> TODO because smart
                    points_for_triangulation.append(point_on_surface(region, labels))
                except:
                    traceback.print_exc()
                    print('no vertices detected for this cell2! --> need fix the algorithm')
                    # --> indeed it fails at detecting cells at the edges!!! -> need fix that --> see how I was doing that in TA and do the same here
                    continue

                try:
                    # there is a bug in there dunno where
                    triangles = triangulate(np.asarray(points_for_triangulation), triangulated_object_ID=cell_ID, labels_to_check=clone_id) # need filter the triangulation for concave shapes of clones which is very frequent

                    # make it measure the 3D area of the clone --> TODO
                    print('area cell', cell_ID, compute_3D_surfacearea(triangles, heightmap=heightmap)) # --> ça marche enfin on dirait! --> cool

                    plot_triangulation(triangles, out)
                except:
                    traceback.print_exc()
                    print('failed for cell2', region.area,
                          region.bbox)  # area 64 but not enough vertices --> I probably have a bug in my vertices code --> TODO --> fix that --> most likely at the edges
                    # TODO fix the algo

                # print(triangles)
                # break
        plt.imshow(out)
        plt.show()

        # plt.imshow(labels)
        # plt.show()

        # tt marche --> plutot cool en fait





# TODO add glob to loadlist to load generic lists --> good idea I think and offer various sorting algos including natsort
# a tool to load lists that is smarter than the TA list loader

import os
import traceback
from functools import partial

from natsort import natsorted
# from epyseg.img import _transfer_voxel_size_metadata
from epyseg.tools.early_stopper_class import early_stop
from epyseg.tools.logger import TA_logger # logging
import glob
import shutil

from epyseg.utils.commontools import execute_chained_functions_and_save_as_tiff

logger = TA_logger()

# if ouput name is
# if save is True then save
def create_list(input_folder, save=False, output_name=None, extensions=['*.tif', '*.tiff', '*.jpg', '*.jpeg', '*.png', '*.bmp', '*.tga', '*.lif'], sort_type='natsort'):
    lst = []
    if not extensions:
        logger.error('Extensions are not defined --> list cannot be created')
        return lst
    path = input_folder
    if not path.endswith('/') and not path.endswith(
            '\\') and not '*' in path and not os.path.isfile(path):
        path+='/'
    # print('corrected path', path)
    if not '*' in path:
        for ext in extensions:
            if not ext.startswith('*'):
                ext='*'+ext
            lst += glob.glob(path + ext)
    else:
        lst+=glob.glob(path)
    if sort_type=='natsort':
        lst=natsorted(lst)
    if save:
        if output_name is None:
            output_name = os.path.join(input_folder,'list.lst')
        save_list_to_file(lst,output_name)
    return lst


# this allows to create a sublist to resume the analysis, e.g. start over in case of failure
def resume_list_on_value(input_list, file_to_resume_on):
    index = input_list.index(file_to_resume_on)
    if index>0:
        return input_list[index:]


def loadlist(txtfile, always_prefer_local_directory_if_file_exists=True, filter_existing_files_only=False, smart_local_folder_detection=True, skip_hashtag_commented_lines=True, skip_empty_lines=True):
    if not os.path.exists(txtfile) or not os.path.isfile(txtfile):
        if not '*' in txtfile and not os.path.isdir(txtfile):
            return None
        else:
            # try creating a list
            # print("here")
            lst = create_list(txtfile, save=False)
    else:
        with open(txtfile) as f:
            if not skip_hashtag_commented_lines:
                lst = [line.rstrip() for line in f]
            else:
                # exclude commented lines anyways they will make the software crash...
                lst = [line.rstrip() for line in f if not line.rstrip().startswith('#')]
        if skip_empty_lines:
            # excludes lines containing spaces only/empty lines (this also fixes a bug that relative path cannot be found if the file contains an empty line), so altogether I think this should be on by default.
            lst = [line.rstrip() for line in lst if not line.strip() == '']

    if always_prefer_local_directory_if_file_exists:
        logger.debug('Files located in the .lst/.txt file containing folder will be preferred, if they exist.')
        # this should probably be improved for complex dir structure I would need to find the common root of all paths and apply some voodoo to keep this --> think about it for future dvpt
        list_file_directory = os.path.dirname(txtfile)
        if not smart_local_folder_detection:
            lst = [os.path.join(list_file_directory, os.path.basename(line)) if os.path.isfile(os.path.join(list_file_directory, os.path.basename(line))) else line for line in lst]
        else:
            # find common root path and apply local path to common path --> much better list opening than above (because supports sub folders too), smarter code than what I had done in TA
            common_path = os.path.commonprefix(lst)
            # print('common_path', common_path)
            if not os.path.isdir(common_path):
                common_path = os.path.dirname(common_path)
            # print(os.path.isdir(common_path))
            # print('common_path',common_path)
            if common_path != '':
                relative_paths = [os.path.relpath(path, common_path) for path in lst]
                lst = [os.path.join(list_file_directory, rel) if os.path.isfile(os.path.join(list_file_directory, rel)) else line for rel, line in zip(relative_paths, lst)]

    if filter_existing_files_only:
        logger.debug('Checking list for existing files')
        # check whether files exist otherwise skip
        lst = [line for line in lst if os.path.isfile(line)]

    return lst

# TODO maybe do
# this class will inject files in TA folder --> can be used to easily reinject deep learning generated files into TA folders

def TA_smart_pairing_list(TA_list, corresponding_list_of_files_to_inject_or_folder, stop_on_error=False):
    if TA_list.lower().endswith('.txt') or TA_list.lower().endswith('.lst'):
        TA_list = loadlist(TA_list)




    if os.path.isdir(corresponding_list_of_files_to_inject_or_folder):
        # assume same name as for input
        # corresponding_list_of_files_to_inject_or_folder
        matching_list = []
        for file in TA_list:
            filename0_without_path = os.path.basename(file)
            # filename0_without_ext = os.path.splitext(filename0_without_path)[0] # TODO maybe allow the user to change the ext some day --> can inject text files or alike then!!!
            matching_list.append(os.path.join(corresponding_list_of_files_to_inject_or_folder, filename0_without_path))
    elif corresponding_list_of_files_to_inject_or_folder.lower().endswith('.txt') or corresponding_list_of_files_to_inject_or_folder.lower().endswith('.lst'):
        matching_list =  loadlist(corresponding_list_of_files_to_inject_or_folder)

    # try zipping and print correspondance and if fails try smart matching --> but warn maybe
    if len(matching_list) != len(TA_list):
        if stop_on_error:
            print('input lists/folders don\'t match', len(TA_list), len(matching_list))
            return
        else:
            # try some rescue
            # if there is a match in name --> just use that
            # TODO implement that
            # maybe use that for max proj and alike stuff
            # use classical matching patterns and exclude extension --> TODO
            # do a matching without repick
            # can allow to macth with generic names added as prefix of suffix before extension such as proj_, max_,  _test,...

            # TODO implement that --> really useful for real TA but ok for now
            pass

    # print(matching_list)
    # print(list(zip(matching_list, TA_list)))


    zipped = list(zip(matching_list, TA_list))
    return zipped

def TA_injector(TA_list, corresponding_list_of_files_to_inject_or_folder, generic_TA_name_for_injected=None, stop_on_error=True, move_file=False, simulate_only = False):
    # if folder assume same name without the extension and inject it

    zipped = TA_smart_pairing_list(TA_list,corresponding_list_of_files_to_inject_or_folder)
    if zipped is None: #
        # nothing todo --> maybe return an error
        return
    # print(matching_things)

    for matching_things in zipped:
        # I always use this --> put it in a generic class because it can be useful
        file_to_copy = matching_things[0]
        file_to_copy_without_path = os.path.basename(file_to_copy)



        file = matching_things[-1]




        filename0_without_path = os.path.basename(file)
        filename0_without_ext = os.path.splitext(filename0_without_path)[0]
        path = os.path.dirname(file)
        destination_folder = os.path.join(path, filename0_without_ext)

        destination_name = os.path.join(destination_folder, file_to_copy_without_path if generic_TA_name_for_injected is None else generic_TA_name_for_injected)

        # make sure the folder exists for the copy
        if not simulate_only:
            os.makedirs(os.path.dirname(destination_folder), exist_ok=True)
        if not move_file:
            print('copy', file_to_copy, '-->', destination_name)
            if simulate_only:
                continue
            shutil.copy2(file_to_copy, destination_name)
        else:
            print('move', file_to_copy, '-->', destination_name)
            if simulate_only:
                continue
            shutil.move(file_to_copy, destination_name)


# creates based on the main TA list a list of files with alternative names and returns it
# NB will only change name if the file with the alternate name exists --> I think this is the behaviour I want but ok
def smart_TA_list(main_list, name_to_search_first, alternative_name_if_name_to_search_first_does_not_exist=None):
    lst = [os.path.join(os.path.splitext(line)[0], name_to_search_first) for line in main_list]
    if alternative_name_if_name_to_search_first_does_not_exist is not None:
        lst = [line if os.path.isfile(line) or not os.path.isfile(os.path.join(os.path.dirname(line), alternative_name_if_name_to_search_first_does_not_exist)) else os.path.join(os.path.dirname(line), alternative_name_if_name_to_search_first_does_not_exist) for line in lst]
    return lst

# files are separated by commas by default
def save_list_to_file(lst, filename, col_separator='\t'):
    # TODO --> maybe add mkdirs...
    if lst is not None:
        with open(filename, 'w') as f:
            for item in lst:
                if not isinstance(item,list) and not isinstance(item, tuple):
                    # TODO maybe do a smarter split if item is itself a list --> then use file sep as a parser
                    f.write("%s\n" % item)
                else:
                    # if item is a list save it as a csv or tab or any other user defined separated text
                    f.write("%s\n" % col_separator.join([str(i) for i in item]))
        return lst
    else:
        raise ValueError('Empty list --> cannot save it to a file')
        return lst

# try TODO a code that transfers metadata in files

# TODO peut etre faire un list executor --> that runs a code for all the files of a list and maybe say whether can be MTed of not
# --> maybe convert this to a class --> think about it for future dev of pyTA the future TA?

# this gives the optimal folder name for saving images without requiring other stuff, from various models --> use that for ensemble tools
def smart_output_folder_name(lst):
    # gets the smartest name for the output folder can be used for predict --> and does not require to store model name as a txt file
    # get common path then replace / or \ by _ to get a smart and meaningful minimal name even when models change!!
    common_path = os.path.commonprefix(lst)
    # print('common_path', common_path)
    if not os.path.isdir(common_path):
        common_path = os.path.dirname(common_path)

    print(common_path)

    if common_path == '':
        return None

    smart_name = [os.path.splitext(name.replace(common_path,'').replace('/','_').replace('\\','_'))[0] for name in lst]

    # then remove file extension

    # remove / if name starts with that
    smart_name = [name if not name.startswith('_') else name[1:] for name in smart_name]

    print(smart_name)



    return smart_name

# can create a list of files with the same name across several folders generated by different models for example --> can then be used for ensemble or averaging...
def get_transversal_list(root_path):

    transverse_list =[]
    # this lists all the folders --> easy to get in fact !!!

    # root_path = '/E/Sample_images/Consensus_learning/gray/CARE/'
    folders = os.listdir(root_path)
    folders = [os.path.join(root_path, folder) for folder in folders]
    folders = natsorted(folders)
    # list of all folders within the folder

    files_in_first_folder = os.listdir(folders[0])

    # DO a list of lists

    # list all files in first folder and if present in others then add them
    for file in files_in_first_folder:
        # images = {}
        multifolder_list = []
        for folder in folders:
            final_file = os.path.join(folder, file)
            if os.path.isfile(final_file):
                try:
                    multifolder_list.append(final_file)
                    # img = Img(final_file)
                    # if force_channel is not None:
                    #     if img.has_c():
                    #         img = img[..., force_channel]
                    # folder_name = os.path.basename(os.path.dirname(final_file))
                    # images[folder_name] = img
                    # # print(img.shape)
                    # # print('valid', final_file)
                except:
                    print('invalid file', final_file)
        transverse_list.append(multifolder_list)

    return transverse_list


def list_processor(lst, processing_fn, multithreading=True, progress_callback=None, use_save_execution_if_chained=False, name_processor_function_for_saving=None):
    # process all the images with the function (maybe even chain process them if needed --> think about how to do that)
    from tqdm import tqdm
    import multiprocessing
    from multiprocessing import Pool
    from timeit import default_timer as timer

    start = timer()

    # import platform
    #
    # if platform.system() == "Darwin":
    #     multiprocessing.set_start_method('spawn')

    nb_procs = multiprocessing.cpu_count() - 1
    if nb_procs <= 0:
        nb_procs = 1

    if multithreading:
        print('using', nb_procs, 'processors')

    if isinstance(processing_fn, list):
        # function_to_chain_iterable, parameter, output_file_name, reverse = False
        processing_fn = partial(execute_chained_functions_and_save_as_tiff, function_to_chain_iterable=processing_fn,
                                output_file_name=name_processor_function_for_saving, reverse=False)
        # print(processing_fn)
    if multithreading:
        import gc
        # given that python copies all the content of the memory for multiprocessing, it's probably a good idea to clean it up
        gc.collect()
        with Pool(processes=nb_procs) as pool:
            for i, _ in enumerate(tqdm(pool.imap_unordered(processing_fn, lst), total=len(lst))):
                if early_stop.stop:
                    pool.close()
                    pool.join()
                    return
                if progress_callback is not None:
                    try:
                        from qtpy.QtWidgets import QProgressBar
                        if isinstance(progress_callback, QProgressBar):
                            progress_callback.setValue((i / len(lst)) * 100)
                        else:
                            progress_callback.emit((i / len(lst)) * 100)
                    except:
                        traceback.print_exc()
            # not sure I should put the lines below...
            # https://stackoverflow.com/questions/38271547/when-should-we-call-multiprocessing-pool-join --> maybe still a good idea to keep that
            pool.close()
            pool.join()
    else:
        # for i, _ in enumerate(tqdm(pool.imap_unordered(processing_fn, lst), total=len(lst))):
        for i, lst_elm in enumerate(lst):
            result = processing_fn(lst_elm)
            if early_stop.stop:
                return
            if progress_callback is not None:
                try:
                    from qtpy.QtWidgets import QProgressBar
                    if isinstance(progress_callback, QProgressBar):
                        progress_callback.setValue((i / len(lst)) * 100)
                    else:
                        progress_callback.emit((i / len(lst)) * 100)
                except:
                    traceback.print_exc()

    print('total time', timer() - start)

if __name__ == '__main__':

    if True:
        transversal_list =get_transversal_list('/E/Sample_images/Consensus_learning/gray/CARE/')
        print(transversal_list)
        # for each can do an avg for example

        import sys
        sys.exit(0)


    if True:
        import sys
        models = [
            'CARE_model_trained_MAE_in_epyseg_29-1_50steps_per_epoch_not_outstanding_though/CARE_model-0.h5',  # 0 #####
            'CARE_model_trained_MSE_error_instead_MAE_in_epyseg_29-1/CARE_model-0.h5', #1
            'CARESEG3D_models_trained_on_colab_201214/CARESEG3D_model-0.h5', #2
            'CARESEG_another_normal_training_201216_colab_but_gave_crap/CARESEG_model-0.h5', # 3 # not bad in fact especially when there is an amazing amount of noise because it is an amazing denoiser...
            'CARESEG/CARESEG_model-0.h5', #4
            'CARESEG_first_correct_test/CARESEG_model-0-0-0.h5', #5
            ## 'CARESEG_first_correct_test/CARESEG_model-0-0-4.h5', #6
            'CARESEG_first_successful_retrain_on_CARE_data_very_very_good_surface_proj_but_not_for_all/CARESEG_model-0-0.h5', #7
            # 'CARESEG_first_successful_retrain_on_CARE_data_very_very_good_surface_proj_but_not_for_all/CARESEG_model-0-1.h5', #8
            # 'CARESEG_first_successful_retrain_on_CARE_data_very_very_good_surface_proj_but_not_for_all/CARESEG_model-0-2.h5', #9
            # 'CARESEG_first_successful_retrain_on_CARE_data_very_very_good_surface_proj_but_not_for_all/CARESEG_model-0-3.h5',#10
            # 'CARESEG_first_successful_retrain_on_CARE_data_very_very_good_surface_proj_but_not_for_all/CARESEG_model-0-4.h5',#11
            'CARESEG_masked_data_and_masked_loss_to_prevent_identity_v1_210107_good/CARESEG_model-0.h5',#12
            'CARESEG_masked_data_and_partial_masked_loss_to_limit_identity_but_not_remove_it_v1_210107_outstanding_on_some_just_good_on_others/CARESEG_model-0.h5',
            #####
            # 13
            'CARESEG_masked_data_and_partial_masked_loss_to_limit_identity_but_not_remove_it_v2_zRoll_stack_210108_quite_good_but_a_bit_low_signal/CARESEG_model-0.h5',
            ##### doesn't seem to require extra black frames and denoiser sucks because it has a lot of noise --> see the best of all
            ##########DO NOT USE NEO######## 'CARESEG_masked_data_and_partial_masked_loss_to_limit_identity_but_not_remove_it_v2_zRoll_stack_210108_quite_good_but_a_bit_low_signal/complete_model_with_3_outputs.h5', # NEO
            # 14
            'CARESEG_my_loss_mae_first_chan_dice_other_chans/CARESEG_model-0.h5',#15
            'CARESEG_RETRAINED_NEW_LOSS_DICE_201215/CARESEG_model-0.h5',#16
            'CARESEG_retrained_normally_201216/CARESEG_model-0.h5',  # 17 #####
            'CARESEG_second_correct_tested_210219_outstanding_denoise/CARESEG_model-0.h5',  # 18 #####
            'CARESEG_second_successful_retrain_on_CARE_data_better_than_previous_need_longer_training_and_more_diverse_one_too/CARESEG_model-0.h5',#19
            'CARESEG_shuffle_n_roll_models_trained_on_colab_201215/CARESEG_model-0.h5',#20
            'CARESEG_test_combined_loss_properly_made_with_reduce_mean_220105/CARESEG_model-0.h5',#21
            'CARESEG_test_combined_mae_jaccard_loss_properly_made_with_reduce_mean_220105_terrible_no_clue_why/CARESEG_model-0.h5',# 22
            'CARESEG_trained_different_losses_for_channels_mae_bce/CARESEG_model-0.h5', # 23
            'CARESEG_trained_on_CARE_data_with_dilation_good/CARESEG_model-0.h5',# 24
            'CARESEG_Zroll_models_trained_on_colab_201215/CARESEG_model-0.h5',# 25
            'CARE_trained_on_its_data_and_mine_not_working_on_its/CARE_model-0.h5',# 26
            'CARE_training_on_its_own_data_with_CARESEG/CARESEG_model-0-0.h5',# 27
            # 'CARE_training_on_its_own_data_with_CARESEG/CARESEG_model-0-4.h5',# 28
            'retrained_CARE_my_soft_fixed_normalization_not_outstanding_but_improving/CARE_model-0.h5',# 29
            'CARE_finally_working_retrain_with_CARE_normalization/CARE_model-0.h5',  # 30
            'CARE_models_trained_on_colab_201214/CARE_model-0.h5',  # 31 #####
            'CARE_masked_data_and_partial_masked_loss_to_limit_identity_but_not_remove_it_v1_210120_not_outstanding_CARESEG_is_better/CARE_model-0.h5',# 32
        ]

        for idx, model in enumerate(models):
            models[idx] = os.path.join('/E/models/my_model/bckup/', model)

        print(models)

        smart_output_folder_name(models)

        # get smart_output_folder_name

        sys.exit(0)


    if False:
        import sys

        # THIS CODE COPIES METADATA FROM ONE FILE TO ANOTHER
        # CAN I USE LEICA DATA TODO STUFF
        # I may still need a time file at least for a while until I change my other tools but that could be useful still
        # TODO --> make this its own function or class !!!!
        # do a generic executer within this class and say whether it can be Mted or not ... --> Could be useful
        list_file = [
            '/E/Sample_images/sample_images_denoise_manue/210402_EcadKI_mel_female_26hAPF_pupae_from_old_tube/predict/predict_model_nb_0_not_best_but_very_good_still/list.lst']  # done 3D
        list_file.append(
            '/E/Sample_images/sample_images_denoise_manue/210324_ON_suz_22h45_armGFP_line2/predict/predict_model_nb_0/mini_list.lst')
        list_file.append(
            '/E/Sample_images/sample_images_denoise_manue/fullset_Manue/Ovipositors/21-03/GT/predict/predict_model_nb_3/list.lst')
        list_file.append(
            '/E/Sample_images/sample_images_denoise_manue/210312_armGFP_line2_suz_39h30APF/predict/predict_model_nb_0/list.lst')

        list_file.append(
            '/E/Sample_images/sample_images_denoise_manue/fullset_Manue/Ovipositors/25-1/GT/predict/predict_model_nb_3/list.lst')
        list_file.append(
            '/E/Sample_images/sample_images_denoise_manue/fullset_Manue/Ovipositors/070219/predict/predict_model_nb_0/list.lst')
        list_file.append(
            '/E/Sample_images/sample_images_denoise_manue/fullset_Manue/Ovipositors/20-2/predict/predict_model_nb_0/list.lst')
        list_file.append(
            '/E/Sample_images/sample_images_denoise_manue/fullset_Manue/Ovipositors/210219/predict/predict_model_nb_0/list.lst')
        list_file.append(
            '/E/Sample_images/sample_images_denoise_manue/fullset_Manue/Ovipositors/27-02/predict/predict_model_nb_0/list.lst')
        list_file.append(
            '/E/Sample_images/sample_images_denoise_manue/210409_EcadKI_ovipo_female_around_40hAPF/predict/predict_model_nb_4/list.lst')
        # list_file = []
        list_file.append('/E/Sample_images/sample_images_denoise_manue/210121_armGFP_suz_line2_47h30_APF/predict/before_crash/predict_model_nb_0/mini_list.lst')

        # list_file = []
        list_file.append(
            '/E/Sample_images/sample_images_denoise_manue/201104_armGFP_different_lines_tila/predict/predict_model_nb_0/mini_list.lst')  # TODO

        # HACK FOR THE LAST BECAUSE IT IS INCORRECT!!!
        # zipped = TA_smart_pairing_list('/E/Sample_images/sample_images_denoise_manue/201104_armGFP_different_lines_tila/predict/predict_model_nb_0/registered/splitted/list2_matching.lst', '/E/Sample_images/sample_images_denoise_manue/201104_armGFP_different_lines_tila/list1.lst')

        # map 2 lists
        for lst in list_file:

            parent_folder = heightmap_folder = lst.split('/predict/')[0]
            # print(os.path.exists(heightmap_folder))
            if not os.path.exists(parent_folder):
                print('error', parent_folder)

            # print(parent_folder)

            # zipped = TA_smart_pairing_list(lst, parent_folder)

            # HACK FOR THE LAST BECAUSE IT IS INCORRECT!!!
            # zipped = TA_smart_pairing_list('/E/Sample_images/sample_images_denoise_manue/201104_armGFP_different_lines_tila/predict/predict_model_nb_0/registered/splitted/list2_matching.lst', '/E/Sample_images/sample_images_denoise_manue/201104_armGFP_different_lines_tila/list1.lst')
            zipped = TA_smart_pairing_list(lst, parent_folder)
            # print(zipped)
            for files in zipped:
                print('transferring metadata from', files[0], 'to', files[1])
                _transfer_voxel_size_metadata(files[0], files[1])#.replace('.tif','.lsm')
                # pass
        sys.exit(0)

    if True:
        # TODO add glob


        import sys
        # copy and inject height maps to the file
        # TA_injector
        simulate_only = False
        generic_TA_name = 'height_map.tif'

        list_file = ['/E/Sample_images/sample_images_denoise_manue/210402_EcadKI_mel_female_26hAPF_pupae_from_old_tube/predict/predict_model_nb_0_not_best_but_very_good_still/list.lst']  # done 3D

        list_file.append('/E/Sample_images/sample_images_denoise_manue/210324_ON_suz_22h45_armGFP_line2/predict/predict_model_nb_0/mini_list.lst')
        list_file.append('/E/Sample_images/sample_images_denoise_manue/fullset_Manue/Ovipositors/21-03/GT/predict/predict_model_nb_3/list.lst')
        list_file.append('/E/Sample_images/sample_images_denoise_manue/210312_armGFP_line2_suz_39h30APF/predict/predict_model_nb_0/list.lst')
        list_file.append('/E/Sample_images/sample_images_denoise_manue/fullset_Manue/Ovipositors/25-1/GT/predict/predict_model_nb_3/list.lst')
        list_file.append('/E/Sample_images/sample_images_denoise_manue/fullset_Manue/Ovipositors/070219/predict/predict_model_nb_0/list.lst')
        list_file.append('/E/Sample_images/sample_images_denoise_manue/fullset_Manue/Ovipositors/20-2/predict/predict_model_nb_0/list.lst')
        list_file.append('/E/Sample_images/sample_images_denoise_manue/fullset_Manue/Ovipositors/210219/predict/predict_model_nb_0/list.lst')
        list_file.append('/E/Sample_images/sample_images_denoise_manue/fullset_Manue/Ovipositors/27-02/predict/predict_model_nb_0/list.lst')
        list_file.append('/E/Sample_images/sample_images_denoise_manue/210409_EcadKI_ovipo_female_around_40hAPF/predict/predict_model_nb_4/list.lst')
        list_file.append(
            '/E/Sample_images/sample_images_denoise_manue/210121_armGFP_suz_line2_47h30_APF/predict/before_crash/predict_model_nb_0/mini_list.lst')
        # TODO hack the last one
        # list_file = []
        # list_file.append('/E/Sample_images/sample_images_denoise_manue/201104_armGFP_different_lines_tila/predict/predict_model_nb_0/registered/splitted/mini_list.lst')
        list_file.append('/E/Sample_images/sample_images_denoise_manue/201104_armGFP_different_lines_tila/predict/predict_model_nb_0/mini_list.lst')

        for lst in list_file:
            # TA_injector('/E/Sample_images/sample_images_denoise_manue/210402_EcadKI_mel_female_26hAPF_pupae_from_old_tube/predict/predict_model_nb_0_not_best_but_very_good_still/list.lst', '/E/Sample_images/sample_images_denoise_manue/210402_EcadKI_mel_female_26hAPF_pupae_from_old_tube/predict/height_map_test1', simulate_only=simulate_only, generic_TA_name_for_injected=generic_TA_name)
            heightmap_folder = lst.split('/predict/')[0]+'/predict/height_maps/'
            # print(os.path.exists(heightmap_folder))
            if not os.path.exists(heightmap_folder):
                print(heightmap_folder)
            # TA_injector(list, '/E/Sample_images/sample_images_denoise_manue/210402_EcadKI_mel_female_26hAPF_pupae_from_old_tube/predict/height_map_test1', simulate_only=simulate_only, generic_TA_name_for_injected=generic_TA_name)
            TA_injector(lst, heightmap_folder, simulate_only=simulate_only, generic_TA_name_for_injected=generic_TA_name)
        sys.exit(0)

    # TODO transform this into tests (just create some tmp files)
    print(loadlist('/E/Sample_images/sample_images_denoise_manue/200722_armGFP_suz_ON_47hAPF/predict/predict_model_nb_4/list.lst'))
    print(loadlist('/E/Sample_images/sample_images_PA/trash_test_mem/mini/list.lst'))
    print(loadlist('/E/Sample_images/sample_images_PA/trash_test_mem/mini/list.lst', always_prefer_local_directory_if_file_exists=False))
    print(loadlist('/E/Sample_images/sample_images_PA/trash_test_mem/mini/list.lst', always_prefer_local_directory_if_file_exists=False, filter_existing_files_only=True))
    print(loadlist('/E/Sample_images/sample_images_PA/trash_test_mem/mini/list.lst', always_prefer_local_directory_if_file_exists=True, filter_existing_files_only=True))
    print(loadlist('/E/Sample_images/sample_images_PA/trash_test_mem/mini/list_smart.lst', always_prefer_local_directory_if_file_exists=True, smart_local_folder_detection=False))
    print(loadlist('/E/Sample_images/sample_images_PA/trash_test_mem/mini/list_smart.lst',                  always_prefer_local_directory_if_file_exists=True))
    print(loadlist('/E/Sample_images/sample_images_PA/trash_test_mem/mini/list_smart.lst')) # this is the default behavior and I do really love it
    print(loadlist('/E/Sample_images/sample_images_PA/trash_test_mem/mini/list_smart.lst', always_prefer_local_directory_if_file_exists=True, smart_local_folder_detection=False, filter_existing_files_only=True))

    tmp_lst = loadlist('/E/Sample_images/sample_images_PA/trash_test_mem/mini/list.lst', always_prefer_local_directory_if_file_exists=True, filter_existing_files_only=True)
    print(smart_TA_list(tmp_lst, 'tutu.tif'))
    print(smart_TA_list(tmp_lst, 'handCorrection.tif', alternative_name_if_name_to_search_first_does_not_exist=None))
    print(smart_TA_list(tmp_lst, 'handCorrection.tif', alternative_name_if_name_to_search_first_does_not_exist='handCorrection.png'))
    # I think that works as I want
    print(smart_TA_list(tmp_lst, 'handCorrection2.png', alternative_name_if_name_to_search_first_does_not_exist='handCorrection.png'))

    save_list_to_file(smart_TA_list(tmp_lst, 'handCorrection.tif', alternative_name_if_name_to_search_first_does_not_exist='handCorrection.png'), '/E/Sample_images/sample_images_PA/trash_test_mem/mini/list_masks.lst')

    print('final list',create_list('/E/Sample_images/sample_images_denoiseg/train/raw/', save=True))



from matplotlib import pyplot as plt
from epyseg.deeplearning.augmentation.generators.data import DataGenerator
from epyseg.deeplearning.augmentation.generators.meta import MetaGenerator
import numpy as np
from epyseg.tools.logger import TA_logger # logging

logger = TA_logger()


# MINIMAL_AUGMENTATIONS = [{'type': None}, {'type': None},{'type': None}, {'type': None}, {'type': 'zoom'}, {'type': 'blur'}, {'type': 'translate'}, {'type': 'rotate'}]
# added intensity shifts to the minimal augmentation --> should make it more robust for masking
# {'type': None},

# TODO recode all of those... --> this is very depreacted and really need add the new graded intensity stuff because it's likely to strongly improve the efficiency of the model
# TODO allow several modifs to be applied at the same time ...

# I have recently modified the augmentation so that it takes into account the elastic deformation which is very good training I assume
MINIMAL_AUGMENTATIONS = [{'type': None}, {'type': None},{'type': None},  {'type': 'zoom'}, {'type': 'blur'}, {'type': 'translate'}, {'type': 'rotate'},{'type': 'random_intensity_gamma_contrast'}, {'type': 'intensity'}, {'type': 'random_intensity_gamma_contrast'}, {'type': 'intensity'}, {'type':'elastic'}, {'type':'elastic'}, {'type': 'graded_intensity_modification'}, {'type': 'graded_intensity_modification'}]

# MINIMAL_AUGMENTATIONS_WITH_ELASTIC= [{'type': None}, {'type': None},{'type': None}, {'type': 'zoom'}, {'type': 'blur'}, {'type': 'translate'}, {'type': 'rotate'},{'type': 'random_intensity_gamma_contrast'}, {'type': 'intensity'}, {'type': 'random_intensity_gamma_contrast'}, {'type': 'intensity'}, {'type':'elastic'}, {'type':'elastic'}, {'type':'elastic'}]

ALL_AUGMENTATIONS_BUT_INVERT_AND_HIGH_NOISE = [{'type': None}, {'type': None}, {'type': 'zoom'}, {'type': 'blur'},
                                               {'type': 'translate'}, {'type':'elastic'},
                                               {'type': 'shear'}, {'type': 'flip'}, {'type': 'rotate'},
                                               {'type': 'low noise'}, {'type': 'high noise'}, {'type': 'stretch'}]

ALL_AUGMENTATIONS_BUT_INVERT = [{'type': None}, {'type': None}, {'type': 'zoom'}, {'type': 'blur'},
                                {'type': 'translate'}, {'type':'elastic'},
                                {'type': 'shear'}, {'type': 'flip'}, {'type': 'rotate'}, {'type': 'low noise'},
                                {'type': 'high noise'}, {'type': 'stretch'}]

ALL_AUGMENTATIONS_BUT_INVERT_AND_NOISE = [{'type': None}, {'type': 'zoom'}, {'type': 'blur'},
                                          {'type': 'translate'}, {'type': 'shear'}, {'type':'elastic'},
                                          {'type': 'flip'}, {'type': 'rotate'}, {'type': 'stretch'},
                                          {'type': 'rotate (interpolation free)'},
                                          {'type': 'rotate (interpolation free)'},
                                          {'type': 'rotate (interpolation free)'}]

ALL_AUGMENTATIONS = [{'type': None}, {'type': None}, {'type': 'zoom'}, {'type': 'blur'}, {'type': 'translate'},
                     {'type': 'shear'}, {'type': 'flip'}, {'type': 'rotate'}, {'type': 'invert'}, {'type': 'low noise'},
                     {'type': 'high noise'}, {'type': 'stretch'}, {'type':'elastic'}]

ALL_AUGMENTATIONS_BUT_HIGH_NOISE = [{'type': None}, {'type': None}, {'type': 'zoom'}, {'type': 'blur'},
                                    {'type': 'translate'},
                                    {'type': 'shear'}, {'type': 'flip'}, {'type': 'rotate'}, {'type': 'invert'},
                                    {'type': 'low noise'}, {'type': 'stretch'}]

STRETCHED_AUG_EPITHELIA = [{'type': None}, {'type': 'stretch'}, {'type': 'stretch'},
                       {'type': 'stretch'}, {'type': 'invert'}, {'type': 'flip'}, {'type': 'translate'},
                       {'type': 'zoom'}, {'type': 'blur'}, {'type': 'shear'}, {'type': 'rotate'}, {'type': 'low noise'}]

STRETCHED_AUG_EPITHELIA_2 = [{'type': None}, {'type': None}, {'type': None}, {'type': None}, {'type': 'stretch'}, {'type': 'stretch'},
                       {'type': 'stretch'}, {'type': 'invert'},{'type': 'invert'},{'type': 'invert'},{'type': 'invert'}, {'type': 'flip'}, {'type': 'translate'},
                       {'type': 'zoom'}, {'type': 'blur'}, {'type': 'shear'}, {'type': 'rotate'}, {'type': 'low noise'}]

STRETCHED_AUG_EPITHELIA_3 = [{'type': None}, {'type': None}, {'type': None}, {'type': None}, {'type': 'stretch'}, {'type': 'stretch'},
                       {'type': 'stretch'}, {'type': 'flip'}, {'type': 'translate'},
                       {'type': 'zoom'}, {'type': 'blur'}, {'type': 'shear'}, {'type': 'rotate'}, {'type': 'low noise'},{'type': 'rotate (interpolation free)'}, {'type': 'rotate (interpolation free)'}, {'type': 'rotate (interpolation free)'}]

STRETCHED_AUG_EPITHELIA_4 = [{'type': None}, {'type': None}, {'type': 'stretch'}, {'type': 'stretch'},
                       {'type': 'stretch'}, {'type': 'flip'}, {'type': 'translate'},{'type': 'flip'},  {'type': 'zoom'}, {'type': 'shear'}, {'type': 'rotate'}, {'type': 'rotate'}, {'type': 'rotate'}, {'type': 'rotate'},
                       {'type': 'zoom'}, {'type': 'blur'}, {'type': 'shear'}, {'type': 'rotate'}, {'type': 'low noise'},{'type': 'rotate (interpolation free)'}, {'type': 'rotate (interpolation free)'}, {'type': 'rotate (interpolation free)'}]


TRAINING_FOR_BEGINNING_LITTLE_INTERPOLATION =  [{'type': 'rotate (interpolation free)'}, {'type': 'rotate (interpolation free)'}, {'type': 'rotate (interpolation free)'}, {'type': None}, {'type': 'flip'}, {'type': 'translate'}, {'type': 'blur'}]

NO_AUGMENTATION = [{'type': None}]

TEST_AUGMENTATION = [{'type': 'graded_intensity_modification'}]#[{'type': 'elastic'}]#[{'type': 'invert'}]

SAFE_AUGMENTATIONS_FOR_SINGLE_PIXEL_WIDE = [{'type': None}, {'type': 'blur'}, {'type': 'translate'}, {'type': 'flip'}]

SAFE_AUGMENTATIONS_FOR_SINGLE_PIXEL_WIDE_PLUS_INVERT_AND_NOISE = [{'type': None}, {'type': 'blur'},
                                                                  {'type': 'translate'}, {'type': 'flip'},
                                                                  {'type': 'invert'}, {'type': 'low noise'}]


class MetaAugmenter:

    def __init__(self, inputs=None, outputs=None, output_folder=None, input_shape=(None, None, None, 1),
                 output_shape=(None, None, None, 1), input_channel_of_interest=None, output_channel_of_interest=None,
                 input_channel_reduction_rule='copy channel of interest to all channels',
                 input_channel_augmentation_rule='copy channel of interest to all channels',
                 output_channel_reduction_rule='copy channel of interest to all channels',
                 output_channel_augmentation_rule='copy channel of interest to all channels',
                 augmentations=None, treat_some_inputs_as_outputs=False,
                 crop_parameters=None, mask_dilations=None, infinite=False,
                 default_input_tile_width=128, default_input_tile_height=128,
                 default_output_tile_width=128, default_output_tile_height=128,
                 keep_original_sizes=False,
                 input_normalization={'method': 'Rescaling (min-max normalization)', 'range': [0, 1],
                                      'individual_channels': True},
                 output_normalization={'method': 'Rescaling (min-max normalization)', 'range': [0, 1],
                                       'individual_channels': True},
                 validation_split=0, test_split=0,
                 shuffle=True, clip_by_frequency=None, is_predict_generator=False, overlap_x=0, overlap_y=0,
                 batch_size=None, batch_size_auto_adjust=False, invert_image=False, input_bg_subtraction=None, pre_processing_input__or_output=None,  create_epyseg_style_output=None, remove_n_border_mask_pixels=None,
                 is_output_1px_wide=False, rebinarize_augmented_output=False,
                 rotate_n_flip_independently_of_augmentation=False,
                 mask_lines_and_cols_in_input_and_mask_GT_with_nans=None, # should be 'id' or 'noid' and requires a custom loss and metrics --> can only be applied with some losses
                 z_frames_to_add=None,
                 **kwargs):

        self.augmenters = []

        self.inputs = inputs
        self.outputs = outputs
        self.output_folder = output_folder
        self.input_shape = input_shape
        self.output_shape = output_shape
        self.input_channel_of_interest = input_channel_of_interest
        self.output_channel_of_interest = output_channel_of_interest

        self.input_channel_reduction_rule = input_channel_reduction_rule
        self.input_channel_augmentation_rule = input_channel_augmentation_rule
        self.output_channel_reduction_rule = output_channel_reduction_rule
        self.output_channel_augmentation_rule = output_channel_augmentation_rule

        self.augmentations = augmentations
        self.treat_some_inputs_as_outputs= treat_some_inputs_as_outputs
        self.crop_parameters = crop_parameters
        self.batch_size = batch_size
        self.batch_size_auto_adjust = batch_size_auto_adjust
        self.invert_image = invert_image
        self.input_bg_subtraction = input_bg_subtraction
        self.pre_processing_input__or_output= pre_processing_input__or_output
        self.create_epyseg_style_output=create_epyseg_style_output
        self.remove_n_border_mask_pixels = remove_n_border_mask_pixels
        self.is_output_1px_wide = is_output_1px_wide
        self.rebinarize_augmented_output = rebinarize_augmented_output
        self.rotate_n_flip_independently_of_augmentation = rotate_n_flip_independently_of_augmentation
        self.mask_lines_and_cols_in_input_and_mask_GT_with_nans = mask_lines_and_cols_in_input_and_mask_GT_with_nans
        self.z_frames_to_add = z_frames_to_add
        self.mask_dilations = mask_dilations
        self.infinite = infinite
        self.default_input_tile_width = default_input_tile_width
        self.default_input_tile_height = default_input_tile_height
        self.default_output_tile_width = default_output_tile_width
        self.default_output_tile_height = default_output_tile_height
        self.keep_original_sizes = keep_original_sizes
        self.input_normalization = input_normalization
        self.output_normalization = output_normalization
        self.validation_split = validation_split
        self.test_split = test_split
        self.shuffle = shuffle
        self.clip_by_frequency = clip_by_frequency
        self.is_predict_generator = is_predict_generator
        self.overlap_x = overlap_x
        self.overlap_y = overlap_y

        if inputs is not None:
            for i, inp in enumerate(inputs):
                if outputs is not None:
                    cur_output = outputs[i]
                else:
                    cur_output = None
                self.augmenters.append(
                    DataGenerator(inputs=inp, outputs=cur_output, output_folder=output_folder, input_shape=input_shape,
                                  output_shape=output_shape, input_channel_of_interest=input_channel_of_interest,
                                  output_channel_of_interest=output_channel_of_interest,
                                  input_channel_reduction_rule=input_channel_reduction_rule,
                                  input_channel_augmentation_rule=input_channel_augmentation_rule,
                                  output_channel_reduction_rule=output_channel_reduction_rule,
                                  output_channel_augmentation_rule=output_channel_augmentation_rule,
                                  augmentations=augmentations,
                                  treat_some_inputs_as_outputs=treat_some_inputs_as_outputs,
                                  crop_parameters=crop_parameters,
                                  mask_dilations=mask_dilations,
                                  infinite=infinite, default_input_tile_width=default_input_tile_width,
                                  default_input_tile_height=default_input_tile_height,
                                  default_output_tile_width=default_output_tile_width,
                                  default_output_tile_height=default_output_tile_height,
                                  keep_original_sizes=keep_original_sizes,
                                  input_normalization=input_normalization,
                                  output_normalization=output_normalization,
                                  validation_split=validation_split, test_split=test_split,
                                  shuffle=shuffle,
                                  clip_by_frequency=clip_by_frequency,
                                  is_predict_generator=is_predict_generator, overlap_x=overlap_x, overlap_y=overlap_y,
                                  invert_image=invert_image, input_bg_subtraction=input_bg_subtraction, pre_processing_input__or_output=pre_processing_input__or_output, create_epyseg_style_output=create_epyseg_style_output, remove_n_border_mask_pixels=remove_n_border_mask_pixels,
                                  is_output_1px_wide=is_output_1px_wide,
                                  rebinarize_augmented_output=rebinarize_augmented_output,
                                  rotate_n_flip_independently_of_augmentation=rotate_n_flip_independently_of_augmentation,
                                  mask_lines_and_cols_in_input_and_mask_GT_with_nans=mask_lines_and_cols_in_input_and_mask_GT_with_nans,
                                  z_frames_to_add = z_frames_to_add
                                  ))

    def _get_significant_parameter(self, local_param, global_param):
        if local_param is not None:
            return local_param
        else:
            return global_param

    def appendDatasets(self, datasets=None, augmentations=None, **kwargs):

        logger.debug('datasets ' + str(datasets))
        logger.debug('augs ' + str(augmentations))

        if datasets is None:
            return

        # parse and handle inputs
        for dataset in datasets:
            fused = {**dataset, 'augmentations': augmentations}

            # print('fused', fused)

            self.append(**fused)

    # TODO add a way to treat input as mask and maybe ouput as input from the point of view of data aug
    def append(self, inputs=None, outputs=None, output_folder=None, input_shape=None, output_shape=None,
               input_channel_of_interest=None, output_channel_of_interest=None,
               input_channel_reduction_rule=None, input_channel_augmentation_rule=None,
               output_channel_reduction_rule=None, output_channel_augmentation_rule=None,
               augmentations=None, treat_some_inputs_as_outputs=None,
               crop_parameters=None, mask_dilations=None, infinite=None,
               default_input_tile_width=None, default_input_tile_height=None, default_output_tile_width=None,
               default_output_tile_height=None, keep_original_sizes=None, input_normalization=None,
               output_normalization=None, validation_split=None, test_split=None,
               shuffle=None, clip_by_frequency=None,
               is_predict_generator=None, overlap_x=None, overlap_y=None, invert_image=None, input_bg_subtraction=None, pre_processing_input__or_output=None, create_epyseg_style_output=None,
               remove_n_border_mask_pixels=None, is_output_1px_wide=None, rebinarize_augmented_output=None,
               rotate_n_flip_independently_of_augmentation=None,mask_lines_and_cols_in_input_and_mask_GT_with_nans=None,
               z_frames_to_add = None,
               **kwargs):

        # print('debug 123', inputs, outputs, self.inputs, self.outputs)
        # inputs and outputs are ok --> why is there a bug then????

        self.augmenters.append(
            DataGenerator(inputs=self._get_significant_parameter(inputs, self.inputs),
                          outputs=self._get_significant_parameter(outputs, self.outputs),
                          output_folder =self._get_significant_parameter(output_folder, self.output_folder),
                          input_shape=self._get_significant_parameter(input_shape, self.input_shape),
                          output_shape=self._get_significant_parameter(output_shape, self.output_shape),
                          input_channel_of_interest=self._get_significant_parameter(input_channel_of_interest,
                                                                                    self.input_channel_of_interest),
                          output_channel_of_interest=self._get_significant_parameter(output_channel_of_interest,
                                                                                     self.output_channel_of_interest),
                          input_channel_reduction_rule=self._get_significant_parameter(input_channel_reduction_rule,
                                                                                       self.input_channel_reduction_rule),
                          input_channel_augmentation_rule=self._get_significant_parameter(
                              input_channel_augmentation_rule, self.input_channel_augmentation_rule),
                          output_channel_reduction_rule=self._get_significant_parameter(output_channel_reduction_rule,
                                                                                        self.output_channel_reduction_rule),
                          output_channel_augmentation_rule=self._get_significant_parameter(
                              output_channel_augmentation_rule, self.output_channel_augmentation_rule),
                          augmentations=self._get_significant_parameter(augmentations, self.augmentations),
                          treat_some_inputs_as_outputs=self._get_significant_parameter(treat_some_inputs_as_outputs, self.treat_some_inputs_as_outputs),
                          crop_parameters=self._get_significant_parameter(crop_parameters, self.crop_parameters),
                          mask_dilations=self._get_significant_parameter(mask_dilations, self.mask_dilations),
                          infinite=self._get_significant_parameter(infinite, self.infinite),
                          default_input_tile_width=self._get_significant_parameter(default_input_tile_width,
                                                                                   self.default_input_tile_width),
                          default_input_tile_height=self._get_significant_parameter(default_input_tile_height,
                                                                                    self.default_input_tile_height),
                          default_output_tile_width=self._get_significant_parameter(default_output_tile_width,
                                                                                    self.default_output_tile_width),
                          default_output_tile_height=self._get_significant_parameter(default_output_tile_height,
                                                                                     self.default_output_tile_height),
                          keep_original_sizes=self._get_significant_parameter(keep_original_sizes,
                                                                              self.keep_original_sizes),
                          validation_split=self._get_significant_parameter(validation_split, self.validation_split),
                          test_split=self._get_significant_parameter(test_split, self.test_split),
                          shuffle=self._get_significant_parameter(shuffle, self.shuffle),
                          clip_by_frequency=self._get_significant_parameter(clip_by_frequency, self.clip_by_frequency),
                          is_predict_generator=self._get_significant_parameter(is_predict_generator,
                                                                               self.is_predict_generator),
                          overlap_x=self._get_significant_parameter(overlap_x, self.overlap_x),
                          overlap_y=self._get_significant_parameter(overlap_y, self.overlap_y),
                          invert_image=self._get_significant_parameter(invert_image, self.invert_image),
                          input_bg_subtraction=self._get_significant_parameter(input_bg_subtraction, self.input_bg_subtraction),
                          pre_processing_input__or_output=self._get_significant_parameter(pre_processing_input__or_output, self.pre_processing_input__or_output),
                          create_epyseg_style_output=self._get_significant_parameter(create_epyseg_style_output, self.create_epyseg_style_output),
                          remove_n_border_mask_pixels=self._get_significant_parameter(remove_n_border_mask_pixels,
                                                                                      self.remove_n_border_mask_pixels),
                          input_normalization=self._get_significant_parameter(input_normalization,
                                                                              self.input_normalization),
                          output_normalization=self._get_significant_parameter(output_normalization,
                                                                               self.output_normalization),
                          is_output_1px_wide=self._get_significant_parameter(is_output_1px_wide,
                                                                             self.is_output_1px_wide),
                          rebinarize_augmented_output=self._get_significant_parameter(rebinarize_augmented_output,
                                                                                      self.rebinarize_augmented_output),
                          rotate_n_flip_independently_of_augmentation=self._get_significant_parameter(rotate_n_flip_independently_of_augmentation,
                                                                                      self.rotate_n_flip_independently_of_augmentation),
                          mask_lines_and_cols_in_input_and_mask_GT_with_nans=self._get_significant_parameter(mask_lines_and_cols_in_input_and_mask_GT_with_nans,
                                                                                      self.mask_lines_and_cols_in_input_and_mask_GT_with_nans),
                          z_frames_to_add=self._get_significant_parameter(z_frames_to_add, self.z_frames_to_add),
                          ))

    def validation_generator(self, infinite=False):
        if infinite:
            while True:
                for orig, label in self._validation_generator(skip_augment=True):
                    # bug fix for recent tensorflow that really needs true and pred to be unpacked if single input and output
                    if len(orig) == 1:
                        orig = orig[0]
                    if len(label) == 1:
                        label = label[0]
                    yield orig, label
        else:
            for orig, label in self._validation_generator(skip_augment=True):
                # bug fix for recent tensorflow that really needs true and pred to be unpacked if single input and output
                if len(orig) == 1:
                    orig = orig[0]
                if len(label) == 1:
                    label = label[0]
                yield orig, label

    def train_generator(self, infinite=False):
        if infinite:
            while True:
                for orig, label in self._train_generator(skip_augment=False):
                    # bug fix for recent tensorflow that really needs true and pred to be unpacked if single input and output
                    if len(orig) == 1:
                        orig = orig[0]
                    if len(label) == 1:
                        label = label[0]
                    yield orig, label
        else:
            for orig, label in self._train_generator(skip_augment=False):
                # bug fix for recent tensorflow that really needs true and pred to be unpacked if single input and output
                if len(orig) == 1:
                    orig = orig[0]
                if len(label) == 1:
                    label = label[0]
                yield orig, label

    def test_generator(self, infinite=False):
        if infinite:
            while True:
                for orig, label in self._test_generator(skip_augment=True):
                    # bug fix for recent tensorflow that really needs true and pred to be unpacked if single input and output
                    if len(orig) == 1:
                        orig = orig[0]
                    if len(label) == 1:
                        label = label[0]
                    yield orig, label
        else:
            for orig, label in self._test_generator(skip_augment=True):
                # bug fix for recent tensorflow that really needs true and pred to be unpacked if single input and output
                if len(orig) == 1:
                    orig = orig[0]
                if len(label) == 1:
                    label = label[0]
                yield orig, label


    def angular_yielder(self, orig, mask, count):
        # mask = self.extra_watershed_mask(mask) # shrink mask to 1 px wide irrespective of transfo
        # NB could do here the generations of the nine stacks --> TODO --> would increase size by 9 but it is a good idea I think
        # can also copy the code of the other stuff

        if count == 0:
            # rot 180
            return np.rot90(orig, 2, axes=(-3, -2)), np.rot90(mask, 2, axes=(-3, -2))

        if count == 1:
            # flip hor
            return np.flip(orig, -2), np.flip(mask, -2)

        if count == 2:
            # flip ver
            return np.flip(orig, -3), np.flip(mask, -3)

        # make it yield the original and the nine versions of it
        # --> TODO
        # ça marche ça me genere les 9 versions du truc dans tous les sens --> probablement ce que je veux --> tt mettre ici
        if count == 3:
            # yield np.rot90(orig, axes=(-3, -2)), np.rot90(mask, axes=(-3, -2))

            # rot 90
            return np.rot90(orig, axes=(-3, -2)), np.rot90(mask, axes=(-3, -2))

        if count == 4:
            # rot 90_flipped_hor or ver
            return np.flip(np.rot90(orig, axes=(-3, -2)), -2), np.flip(np.rot90(mask, axes=(-3, -2)), -2)

        if count == 5:
            # rot 90_flipped_hor or ver
            return np.flip(np.rot90(orig, axes=(-3, -2)), -3), np.flip(np.rot90(mask, axes=(-3, -2)), -3)

        if count == 6:
            # rot 270
            return np.rot90(orig, 3, axes=(-3, -2)), np.rot90(mask, 3, axes=(-3, -2))

    def _train_generator(self, skip_augment, first_run=False):
        train = MetaGenerator(self.augmenters, shuffle=self.shuffle, batch_size=self.batch_size, gen_type='train')
        for out in train.generator(skip_augment, first_run):
            try:
                # # print(len(out))
                # #  that works check that all are there and all are possible otherwise skip
                # # --> need ensure that width = height
                # # need set a parameter to be sure to use it or not and need remove rotation and flip from augmentation list (or not in fact)
                # orig, mask = out
                # augmentations = 7
                # if orig[0].shape[-2] != orig[0].shape[-3]:
                #     augmentations = 3
                # for aug in range(augmentations):
                #     yield self.angular_yielder(orig, mask, aug)
                # yield orig, mask
                yield out
            except GeneratorExit:
                # bug fix for the GeneratorExit error, see https://stackoverflow.com/questions/46542147/elegant-way-for-breaking-a-generator-loop-generatorexit-error
                pass
                break
            except Exception:
                # failed to generate output --> continue
                continue

    def _test_generator(self, skip_augment, first_run=False):
        test = MetaGenerator(self.augmenters, shuffle=False, batch_size=self.batch_size, gen_type='test')
        for out in test.generator(skip_augment, first_run):
            try:
                # # yield out
                # # print(len(out))
                # #  that works check that all are there and all are possible otherwise skip
                # # --> need ensure that width = height
                # # need set a parameter to be sure to use it or not and need remove rotation and flip from augmentation list (or not in fact)
                # orig, mask = out
                # augmentations = 7
                # if orig[0].shape[-2] != orig[0].shape[-3]:
                #     augmentations = 3
                # for aug in range(augmentations):
                #     yield self.angular_yielder(orig, mask, aug)
                # yield orig, mask
                yield out
            except GeneratorExit:
                # bug fix for the GeneratorExit error, see https://stackoverflow.com/questions/46542147/elegant-way-for-breaking-a-generator-loop-generatorexit-error
                pass
                break
            except Exception:
                # failed to generate output --> continue
                continue

    def _validation_generator(self, skip_augment, first_run=False):
        valid = MetaGenerator(self.augmenters, shuffle=self.shuffle, batch_size=self.batch_size, gen_type='valid')
        for out in valid.generator(skip_augment, first_run):
            try:
                # # yield out
                # # print(len(out))
                # #  that works check that all are there and all are possible otherwise skip
                # # --> need ensure that width = height
                # # need set a parameter to be sure to use it or not and need remove rotation and flip from augmentation list (or not in fact)
                # orig, mask = out
                # augmentations = 7
                # if orig[0].shape[-2] != orig[0].shape[-3]:
                #     augmentations = 3
                # for aug in range(augmentations):
                #     yield self.angular_yielder(orig, mask, aug)
                # yield orig, mask
                yield out
            except GeneratorExit:
                # bug fix for the GeneratorExit error, see https://stackoverflow.com/questions/46542147/elegant-way-for-breaking-a-generator-loop-generatorexit-error
                pass
                break
            except Exception:
                # failed to generate output --> continue
                continue

    def predict_generator(self):  # TODO can use datagen for now
        pass

    def __len__(self):
        # returns the nb of datasets
        if not self.augmenters:
            return 0
        return len(self.augmenters)

    # returns the real nb of batches with the current parameters...
    def get_train_length(self, first_run=False):
        # need run the train algo once with real tiled data to get the counts
        train_generator = self._train_generator(skip_augment=True, first_run=first_run)
        nb_batches = 0
        for _, _ in train_generator:
            nb_batches += 1
        return nb_batches

    def get_test_length(self, first_run=False):
        # need run the train algo once with real tiled data to get the counts
        test_generator = self._test_generator(skip_augment=True, first_run=first_run)
        nb_batches = 0
        for _, _ in test_generator:
            nb_batches += 1
        return nb_batches


    def get_validation_length(self, first_run=False):
        # need run the train algo once with real tiled data to get the counts
        validation_generator = self._validation_generator(skip_augment=True, first_run=first_run)
        nb_batches = 0
        for _, _ in validation_generator:
            nb_batches += 1
        return nb_batches


if __name__ == '__main__':
    pass



import setuptools
from epyseg.epygui import __AUTHOR__, __VERSION__, __EMAIL__

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='epyseg',
    version=__VERSION__,
    author=__AUTHOR__,
    author_email=__EMAIL__,
    description='A deep learning based tool to segment epithelial tissues. The epyseg GUI can be uesd to build, train or run custom networks',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/baigouy/EPySeg',
    package_data={'': ['*.md','*.json']}, # include all .md files and json models from the model zoo
    license='BSD',
    include_package_data=True,
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
    ],
    # TODO put this in requirements.txt file and read the file from here --> can have both methods work
    # below are the required files (they will be installed together with epyseg unless the '--no-deps' tag is used)
    install_requires=[
        # "tensorflow>=2.0.0",  # to allow for mac OS X conda support #shall I put 2.3 now
        # "tensorflow-gpu>=2.0.0;platform_system!='Darwin'",
        "tensorflow>=2.3.1", # pb 2.3.1 is not supported in python 3.8 and there is a pb accessing the layers --> I need a fix but maybe ok for now # I have a modulewrapper bug with tf 2.7.1 --> so until I understand it I will rely on that version TODO should I add ;platform_system=='Darwin' ? --> try it # to allow for mac OS X conda support #shall I put 2.3 now # current collab version is 2.8
        # "tensorflow-gpu>=2.3.1;platform_system!='Darwin'", # tensorflow-gpu==2.3.1 # do I still need to exclude macOS??? --> pb is old macs maybe # apparently this line is useless now and there is no point in having it because tensorflow uses the gpu by default if it's there and properly congigured !
        "segmentation-models==1.0.1",
        # "tensorflow-gpu>=2.0.0", # not required ? # make sure it does not install on OS X to prevent crash if does not exist
        "czifile",
        # "h5py", # should be installed with tensorflow gpu so do I need it ??? # probably better to remove it to avoid intsalling erroneous versions that are incompatible with tensorflow... similarly do I really need to have numpy it will be installed with tf anyways??? --> just try
        "Markdown",
        "matplotlib>=3.5.2", #  required to have matplotlib support of pyqt6 --> which will be used by default in the future
        "numpy",
        "numpydoc",
        "Pillow>=8.1.2", # because there are security issues with below
        # for now intsall both but soon just rely on pyqt6 ???
        "PyQt5", # PyQt5==5.15.4    PyQt5-Qt5==5.15.2 # check versions are ok!!!
        "PyQtWebEngine", #PyQtWebEngine==5.15.4 PyQtWebEngine-Qt5==5.15.2
        "PyQt6", # PyQt5==5.15.4    PyQt5-Qt5==5.15.2 # check versions are ok!!! # may require: sudo apt install qt6-base-dev
        "PyQt6-WebEngine", #PyQtWebEngine==5.15.4 PyQtWebEngine-Qt5==5.15.2
        "read-lif",
        "scikit-image>=0.19.3", #scikit-image>=0.18.1 # 0.19.3 is required for applying 3D affine trafo
        "scipy>=1.7.3", # scipy==1.6.3 # 1.7.3 is required for applying 3D affine trafo
        "scikit-learn>=1.0.2",  # for pyTA contour sorting deprecated (remove ?) # 1.0.2 is required for applying 3D affine trafo now
        "tifffile>=2021.11.2", #tifffile==2021.11.2 # ok bug is now fixed --> can keep all the versions of tifffile! # NB apparently there is a bug with the latest version for saving as imageJ (with metadata) --> I may need fix the version to 2021.11.2 # but try it
        "tqdm",
        "natsort",
        "numexpr",
        "urllib3", # for model download
        "qtawesome", # for the TA icons
        "pandas", #pandas==0.24.2n
        "numba", #numba==0.48.0
        "elasticdeform", # a library to further increase the range of data aug
        "roifile", # for support of IJ ROIs soon
        "prettytable", # for SQL preview in pyTA
        "pyperclip", # for pyta lists
        "QtPy>=2.1.0", # from now on this is how I would handle the pyqt/pyside integration
        # "aicsimageio", # new library t oread czi files (or any file maybe ????)--> replace my entire Img library ???
        # "aicspylibczi>=3.0.5", # required for the above to read czi files !!!
        # "pylibczi", # deprecated --> replaced by the 2 above --> DO NOT USE
        # "pylibCZIrw" # --> a python wrapper for libCZI --> this is the Zeiss made stuff so -> SO COOL

        # "parameterized",#required for unit tests
        # "sympy" # TODO add this if I finally use it in EZF
        # six==1.15.0
    ],

    # !python --version # to check python version in colab

    # extras_require = {
    #     'all':  ["tensorflow-gpu>=2.0.0"]
    # },
    # python_requires='>=3.6, <3.9' #  --> could put < 3.10 but then I would have to change tf version to 2.5 and I haven't tested it (see https://www.tensorflow.org/install/pip?hl=fr) --> should try that first --> check the colba version to see if that works or not !!!
    # python_requires='>=3.6, <3.10' # from 16/03/23 tensorflow is now supported by python 3.9 and tf version is 2.11.0 --> could put < 3.10 but then I would have to change tf version to 2.5 and I haven't tested it (see https://www.tensorflow.org/install/pip?hl=fr) --> should try that first --> check the colba version to see if that works or not !!!
    python_requires='>=3.6, <3.11' # from 04/05/23 colab is using python 3.10.11
    # NB colab uses tf 2.8.0 by default (date = 16/02/22) (https://colab.research.google.com/notebooks/tensorflow_version.ipynb#scrollTo=-XbfkU7BeziQ) --> see if my tool works with that then update
    # python version on colab is 3.7.12 --> therefore it is really worth keeping it that way (import sys  sys.version)
)

# to test with testpy !pip install --no-deps -i https://test.pypi.org/simple/ epyseg==0.1.30

# pip3 freeze # the versions have changed --> try with a fresh install just to see if that works

# TODO add svg libs ???
# my versions in case that's needed some day
# should I add keras for seg models ???
# czifile==2019.7.2
# h5py==2.9.0
# Markdown==3.1.1
# matplotlib==3.1.3
# numpy==1.16.4
# numpydoc==0.9.2
# Pillow==7.0.0
# PyQt5==5.13.0
# PyQtWebEngine==5.13.0
# read-lif==0.2.1
# scikit-image==0.16.2
# scipy==1.4.1
# segmentation-models==1.0.1
# tensorflow==2.1.0
# tensorflow-gpu==2.1.0
# tifffile==2020.2.16
# tqdm==4.42.1
# natsort=7.0.1
# numexpr=2.7.1


#!/usr/bin/python
# -'''- coding: utf-8 -'''-

import os
from epyseg.settings.global_settings import set_UI # set the UI to be used py qtpy
set_UI()
import sys

from qtpy import QtCore
from qtpy.QtWidgets import QDialog, QPlainTextEdit, QVBoxLayout, QApplication, QDialogButtonBox, QSpinBox, QLabel


class DenoiseRecursionDialog(QDialog):

    def __init__(self, parent=None, title=None):
        super(DenoiseRecursionDialog, self).__init__(parent)
        if title is not None:
            self.setWindowTitle(title)

        label = QLabel('How many times should the denoiser module be run on the image?\n(Ideally keep this number to 1 and increase only when the output is noisy)\n(typically 4 or 5 is a good value for very noisy images)')
        self.recursion_spinner = QSpinBox()
        self.recursion_spinner.setSingleStep(1)
        self.recursion_spinner.setRange(1, 20)  # 1_000_000 makes no sense but anyway
        self.recursion_spinner.setValue(1)
        layout = QVBoxLayout()
        layout.addWidget(label)
        layout.addWidget(self.recursion_spinner)
        # OK and Cancel buttons
        self.buttons = QDialogButtonBox(
            QDialogButtonBox.Ok | QDialogButtonBox.Cancel,
            QtCore.Qt.Horizontal, self)
        self.buttons.accepted.connect(self.accept)
        self.buttons.rejected.connect(self.reject)
        layout.addWidget(self.buttons)

        # TODO add an ok or cancel button


        self.setLayout(layout)

    @staticmethod
    def get_value(parent=None, title=None):
        dialog = DenoiseRecursionDialog(parent=parent, title=title)
        result = dialog.exec_()
        values = dialog.recursion_spinner.value()
        return (values, result == QDialog.Accepted)

if __name__ == '__main__':
    app = QApplication(sys.argv)
    recursion, ok = DenoiseRecursionDialog.get_value(title="Denoiser options")
    # form.show()
    # text, ok = app.exec_()
    # print(form.get_value())

    # text, ok = QInputDialog.getText(self, 'Text Input Dialog', 'Enter your SQl command:')

    if ok:
        # self.le1.setText(str(text))
        print(recursion)
    else:
        pass

    sys.exit(0)


import os
from epyseg.settings.global_settings import set_UI # set the UI to be used py qtpy
set_UI()
from qtpy import QtCore
from qtpy.QtCore import QPointF, Qt
from qtpy.QtGui import QPainter, QBrush, QPen, QImage, QColor
# from epyseg.draw.shapes.polygon2d import Polygon2D
# from qtpy.Qt
# from qtpy.Qt import (QPaintEngine, QPaintDevice,  QTransform, QBrush)
from epyseg.draw.shapes.polygon2d import Polygon2D
from epyseg.draw.shapes.polyline2d import PolyLine2D

from epyseg.tools.logger import TA_logger
logger = TA_logger()

class Freehand2D(Polygon2D):

    def __init__(self, *args, color=0xFFFF00, opacity=1., stroke=0.65, line_style=None, theta=0, invert_coords=False, fill_color=None, **kwargs):
        super(Freehand2D, self).__init__()
        self.isSet = False
        if len(args) > 0:
            if isinstance(args[0],tuple):
                for i in range(0, len(args)):
                    if invert_coords:
                        self.append(QPointF(args[i][1], args[i][0]))
                    else:
                        self.append(QPointF(args[i][0], args[i][1]))
            else:
                for i in range(0, len(args), 2):
                    if invert_coords:
                        self.append(QPointF(args[i+1], args[i]))
                    else:
                        self.append(QPointF(args[i], args[i+1]))
            self.isSet = True
        self.color = color
        self.fill_color = fill_color
        self.stroke = stroke
        self.opacity = opacity
        self.line_style = line_style
        # rotation
        self.theta = theta

    # def draw(self, painter, draw=True):
    #     if self.color is None and self.fill_color is None:
    #         return
    #
    #     if draw:
    #         painter.save()
    #         painter.setOpacity(self.opacity)
    #         if self.color is not None:
    #             pen = QPen(QColor(self.color))
    #             if self.stroke is not None:
    #                 pen.setWidthF(self.stroke)
    #             if self.line_style is not None:
    #                 if self.line_style in [Qt.SolidLine, Qt.DashLine, Qt.DashDotLine, Qt.DotLine, Qt.DashDotDotLine]:
    #                     pen.setStyle(self.line_style)
    #                 elif isinstance(self.line_style, list):
    #                     pen.setStyle(Qt.CustomDashLine)
    #                     pen.setDashPattern(self.line_style)
    #             painter.setPen(pen)
    #         else:
    #             painter.setPen(Qt.NoPen)  # required to draw something filled without a border
    #         if self.fill_color is not None:
    #             painter.setBrush(QBrush(QColor(self.fill_color)))
    #         polygon_to_draw = self.translated(0, 0)
    #         if self.scale is not None and self.scale != 1:
    #             polygon_to_draw = self.__scaled()
    #
    #         # print('mid rect_to_plot', rect_to_plot)
    #         if self.translation is not None:
    #             # rect_to_plot.setX(rect_to_plot.x()+self.translation.x())
    #             # rect_to_plot.setY(rect_to_plot.y()+self.translation.y())
    #             polygon_to_draw.translate(self.translation.x(), self.translation.y())
    #
    #         if self.theta is not None and self.theta != 0:
    #             painter.translate(polygon_to_draw.center())
    #             painter.rotate(self.theta)
    #             painter.translate(-polygon_to_draw.center())
    #
    #         if self.fill_color is not None:
    #             print('coloring')
    #             brush = QBrush(self.fill_color)
    #             painter.setBrush(brush)
    #
    #         painter.drawPolygon(polygon_to_draw)
    #         painter.restore()
# it's simply a polygon so draw it as such

    # def draw(self, painter, draw=True):
    #     if self.color is None and self.fill_color is None:
    #         return
    #
    #     if draw:
    #         painter.save()
    #         pen = QPen(QColor(self.color))
    #         # pen.setStyle(Qt.SolidPattern)
    #         if self.stroke is not None:
    #             pen.setWidthF(self.stroke)
    #         if self.line_style is not None:
    #             if self.line_style in [Qt.SolidLine, Qt.DashLine, Qt.DashDotLine, Qt.DotLine, Qt.DashDotDotLine]:
    #                 pen.setStyle(self.line_style)
    #             elif isinstance(self.line_style, list):
    #                 pen.setStyle(Qt.CustomDashLine)
    #                 pen.setDashPattern(self.line_style)
    #         painter.setPen(pen)
    #         painter.setOpacity(self.opacity)
    #         polyline_to_draw = self.translated(0, 0)
    #         if self.scale is not None and self.scale != 1:
    #             polyline_to_draw = self.__scaled()
    #         # print('mid rect_to_plot', rect_to_plot)
    #         if self.translation is not None:
    #             # rect_to_plot.setX(rect_to_plot.x()+self.translation.x())
    #             # rect_to_plot.setY(rect_to_plot.y()+self.translation.y())
    #             polyline_to_draw.translate(self.translation.x(), self.translation.y())
    #         # painter.drawPolygon(polygon_to_draw)
    #         if self.theta is not None and self.theta != 0:
    #             painter.translate(polyline_to_draw.boundingRect().center())
    #             painter.rotate(self.theta)
    #             painter.translate(-polyline_to_draw.boundingRect().center())
    #
    #
    #         if self.fill_color is not None:
    #
    #             print('coloring')
    #             brush = QBrush(self.fill_color)
    #             painter.setBrush(brush)
    #
    #         # painter.drawPolyline(polyline_to_draw)
    #         painter.drawPolygon(polyline_to_draw)
    #         # painter.toFillPolygon(polyline_to_draw)
    #         # painter.fill(polyline_to_draw)
    #         painter.restore()


    # def draw(self, painter, draw=True):
    #     if draw:
    #         painter.save()
    #     pen = QPen(QColor(self.color))
    #     if self.stroke is not None:
    #         pen.setWidthF(self.stroke)
    #     painter.setPen(pen)
    #     painter.setOpacity(self.opacity)
    #     if draw:
    #         painter.drawPolyline(self)
    #         painter.restore()



    # def fill(self, painter, draw=True):
    #     if self.fill_color is None:
    #         return
    #     if draw:
    #         painter.save()
    #     painter.setBrush(QBrush(QColor(self.fill_color)))
    #     painter.setOpacity(self.opacity)
    #     if draw:
    #         painter.drawPolyline(self)
    #         painter.restore()
    #
    # # TODO pb will draw the shape twice.... ---> because painter drawpolygon is called twice
    # def drawAndFill(self, painter):
    #     painter.save()
    #     self.draw(painter, draw=False)
    #     self.fill(painter, draw=False)
    #     painter.drawPolyline(self)
    #     painter.restore()

    def setP1(self, point):
        self.append(point)

    def add(self, *args):
        self.append(args[1])
        self.isSet=True

if __name__ == '__main__':
    test = Freehand2D(0, 0, 10, 0, 10, 20, 0, 20, 0, 0)
    print(test.count()) # marche pas --> pas ajouté
    print(test)

    hexagon = Freehand2D()
    print(hexagon)
    hexagon.append(QPointF(10, 20))
    hexagon.append(QPointF(10, 30))
    hexagon.append(QPointF(20, 30))

    print(hexagon)
    print(hexagon.isEmpty())
    print(hexagon.count())
    print(hexagon.stroke)
    print(hexagon.get_stroke_size())
    # ça marche maintenant
    # print(hexagon.)
    # trop cool l'acces aux points
    for point in hexagon:
        print(point)

    print(hexagon.get_points())
    print(hexagon.contains(QPointF(10, 20)))
    print(hexagon.contains(QPointF(10, 21)))

    print(hexagon.isClosed()) #
    hexagon.append(QPointF(10, 20)) # closing the hexagon --> the last and first point should be the same
    print(hexagon.isClosed())  #

    image = QImage('./../data/handCorrection.png')
    painter = QPainter()
    painter.begin(image)
    # painter.setOpacity(0.3);
    painter.drawImage(0, 0, image)
    painter.setPen(QtCore.Qt.blue)
    painter.drawPolygon(hexagon)
    hexagon.opacity = 0.7
    painter.translate(10, 20)
    hexagon.draw(painter) # ça marche pourrait overloader ça avec du svg
    painter.end()

    # painter.save()
    # painter.setCompositionMode(QtGui.QPainter.CompositionMode_Clear)
    # painter.eraseRect(r)
    # painter.restore()

    image.save('./../trash/test_pyQT_draw.png', "PNG");

    #pas mal TODO faire une classe drawsmthg qui dessine n'importe quelle forme que l'on lui passe avec des parametres de couleur, transparence, ...

    # tt marche aps mal ça va très vite

